from gaphas.item import Item

from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.UML.actions.activitynodes import DecisionNodeItem, ForkNodeItem


def test_fork_node_item_implements_item_protocol(diagram):
    fork_node_item = diagram.create(ForkNodeItem)

    assert isinstance(fork_node_item, Item)


def test_decision_node_persistence(diagram, element_factory, saver, loader):
    item = diagram.create(
        DecisionNodeItem, subject=element_factory.create(UML.DecisionNode)
    )

    data = saver()
    loader(data)
    new_diagram = next(element_factory.select(Diagram))
    item = next(new_diagram.select(DecisionNodeItem))

    assert item.combined is None, item.combined


def test_combined_decision_node_persistence(diagram, element_factory, saver, loader):
    item = diagram.create(
        DecisionNodeItem, subject=element_factory.create(UML.DecisionNode)
    )
    merge_node = element_factory.create(UML.MergeNode)
    item.combined = merge_node

    data = saver()
    loader(data)
    new_diagram = next(element_factory.select(Diagram))
    item = next(new_diagram.select(DecisionNodeItem))

    assert item.combined is not None, item.combined
    assert isinstance(item.combined, UML.MergeNode)


def test_fork_node_persistence(diagram, element_factory, saver, loader):
    item = diagram.create(ForkNodeItem, subject=element_factory.create(UML.ForkNode))

    data = saver()
    loader(data)

    new_diagram = next(element_factory.select(Diagram))
    item = next(new_diagram.select(ForkNodeItem))
    assert item.combined is None, item.combined


def test_combined_fork_node_persistence(diagram, element_factory, saver, loader):
    item = diagram.create(ForkNodeItem, subject=element_factory.create(UML.ForkNode))

    merge_node = element_factory.create(UML.JoinNode)
    item.combined = merge_node

    data = saver()
    loader(data)
    new_diagram = next(element_factory.select(Diagram))
    item = next(new_diagram.select(ForkNodeItem))

    assert item.combined is not None, item.combined
    assert isinstance(item.combined, UML.JoinNode)



from gaphor import UML
from gaphor.diagram.drop import drop
from gaphor.UML.classes import PackageItem
from gaphor.UML.interactions import MessageItem
from gaphor.UML.interactions.interactionsconnect import connect_lifelines
from gaphor.UML.recipes import (
    create_association,
    create_extension,
    create_generalization,
)


def test_drop_class(diagram, element_factory):
    klass = element_factory.create(UML.Class)

    drop(klass, diagram, 0, 0)

    assert klass.presentation
    assert klass.presentation[0] in diagram.ownedPresentation


def test_drop_dependency(diagram, element_factory):
    client = element_factory.create(UML.Class)
    supplier = element_factory.create(UML.Class)
    dependency = element_factory.create(UML.Dependency)
    dependency.client = client
    dependency.supplier = supplier

    drop(client, diagram, 0, 0)
    drop(supplier, diagram, 0, 0)
    item = drop(dependency, diagram, 0, 0)

    assert (
        diagram.connections.get_connection(item.head).connected
        is supplier.presentation[0]
    )
    assert (
        diagram.connections.get_connection(item.tail).connected
        is client.presentation[0]
    )


def test_drop_association(diagram, element_factory):
    a = element_factory.create(UML.Class)
    b = element_factory.create(UML.Class)
    association = create_association(a, b)

    drop(a, diagram, 0, 0)
    drop(b, diagram, 0, 0)
    item = drop(association, diagram, 0, 0)

    assert item
    assert diagram.connections.get_connection(item.head).connected is a.presentation[0]
    assert diagram.connections.get_connection(item.tail).connected is b.presentation[0]


def test_drop_generalization(diagram, element_factory):
    a = element_factory.create(UML.Class)
    b = element_factory.create(UML.Class)
    generalization = create_generalization(a, b)

    drop(a, diagram, 0, 0)
    drop(b, diagram, 0, 0)
    item = drop(generalization, diagram, 0, 0)

    assert item
    assert diagram.connections.get_connection(item.head).connected is b.presentation[0]
    assert diagram.connections.get_connection(item.tail).connected is a.presentation[0]


def test_drop_extension(diagram, element_factory):
    metaclass = element_factory.create(UML.Class)
    stereotype = element_factory.create(UML.Stereotype)
    extension = create_extension(metaclass, stereotype)

    drop(metaclass, diagram, 0, 0)
    drop(stereotype, diagram, 0, 0)
    item = drop(extension, diagram, 0, 0)

    assert item
    assert (
        diagram.connections.get_connection(item.head).connected
        is metaclass.presentation[0]
    )
    assert (
        diagram.connections.get_connection(item.tail).connected
        is stereotype.presentation[0]
    )


def test_drop_message(diagram, element_factory):
    a = element_factory.create(UML.Lifeline)
    b = element_factory.create(UML.Lifeline)

    a_item = drop(a, diagram, 0, 0)
    b_item = drop(b, diagram, 0, 0)
    msg_item = diagram.create(MessageItem)
    connect_lifelines(msg_item, a_item, b_item)
    message = msg_item.subject

    item = drop(message, diagram, 0, 0)

    assert item
    assert diagram.connections.get_connection(item.head).connected is a.presentation[0]
    assert diagram.connections.get_connection(item.tail).connected is b.presentation[0]


def test_drop_message_send_connected(diagram, element_factory):
    a = element_factory.create(UML.Lifeline)

    a_item = drop(a, diagram, 0, 0)
    msg_item = diagram.create(MessageItem)
    connect_lifelines(msg_item, send=a_item, received=None)
    message = msg_item.subject

    item = drop(message, diagram, 0, 0)

    assert item
    assert diagram.connections.get_connection(item.head).connected is a.presentation[0]
    assert not diagram.connections.get_connection(item.tail)


def test_drop_message_received_connected(diagram, element_factory):
    b = element_factory.create(UML.Lifeline)

    b_item = drop(b, diagram, 0, 0)
    msg_item = diagram.create(MessageItem)
    connect_lifelines(msg_item, send=None, received=b_item)
    message = msg_item.subject

    item = drop(message, diagram, 0, 0)

    assert item
    assert not diagram.connections.get_connection(item.head)
    assert diagram.connections.get_connection(item.tail).connected is b.presentation[0]


def test_drop_pin(diagram, element_factory):
    action = element_factory.create(UML.Action)
    input_pin = element_factory.create(UML.InputPin)
    input_pin.opaqueAction = action
    output_pin = element_factory.create(UML.OutputPin)
    output_pin.opaqueAction = action

    drop(action, diagram, 0, 0)
    input_item = drop(input_pin, diagram, 0, 0)

    assert input_item
    assert input_item.subject is input_pin
    assert input_item.parent.subject is action

    ouput_item = drop(output_pin, diagram, 0, 0)

    assert ouput_item
    assert ouput_item.subject is output_pin
    assert ouput_item.parent.subject is action


def test_drop_element_on_presentation(diagram, element_factory):
    package_item = diagram.create(
        PackageItem, subject=element_factory.create(UML.Package)
    )
    cls = element_factory.create(UML.Class)

    drop(cls, package_item, 0, 0)
    class_item = cls.presentation[0]

    assert class_item.parent is package_item


def test_drop_element_on_presentation_that_cannot_be_owned(diagram, element_factory):
    package_item = diagram.create(
        PackageItem, subject=element_factory.create(UML.Package)
    )
    action = element_factory.create(UML.Action)

    drop(action, package_item, 0, 0)
    class_item = action.presentation[0]

    assert class_item.parent is None



from pytest_archon import archrule

import gaphor
from gaphor.entrypoint import load_entry_points

GAPHOR_CORE = [
    "gaphor.core*",
    "gaphor.abc",
    "gaphor.asyncio",
    "gaphor.action",
    "gaphor.entrypoint",
    "gaphor.i18n",
    "gaphor.transaction",
    "gaphor.event",
    "gaphor.settings",
]

GLIB = [
    "gi.repository.GLib",
    "gi.repository.GObject",
    "gi.repository.Gio",
]

# Pango is used for text rendering in diagrams
PANGO = [
    "gi.repository.Pango",
    "gi.repository.PangoCairo",
]

UI_LIBRARIES = [
    "gi.repository.Adw",
    "gi.repository.Gdk",
    "gi.repository.Gtk",
]


def test_core_packages():
    (
        archrule("gaphor.core does not depend on the rest of the system")
        .match("gaphor.core*")
        .exclude("*.tests.*")
        .may_import(*GAPHOR_CORE)
        .should_not_import("gaphor*")
        .should_not_import("gi*")
        .check(gaphor, skip_type_checking=True)
    )


def test_diagram_package():
    (
        archrule("Diagrams are part of the core")
        .match("gaphor.diagram*")
        .exclude("*.tests.*")
        .may_import(*GAPHOR_CORE)
        .may_import("gaphor.diagram*")
        .should_not_import("gaphor*")
        .check(gaphor, skip_type_checking=True)
    )

    (
        archrule("GTK dependencies")
        .match("gaphor.diagram*")
        .exclude("*.tests.*")
        .exclude("gaphor.diagram.general.uicomponents")
        .exclude("gaphor.diagram.tools*")
        .exclude("gaphor.diagram.*editors")
        .exclude("gaphor.diagram.*propertypages")
        .exclude("gaphor.diagram.styleeditor")
        .may_import(*PANGO)
        .should_not_import("gi.repository*")
        .check(gaphor, skip_type_checking=True)
    )


def test_services_package():
    (
        archrule("Services only depend on core functionality")
        .match("gaphor.services*")
        .exclude("*.tests.*")
        .may_import(*GAPHOR_CORE, *GLIB)
        .may_import("gaphor.diagram*")
        .may_import("gaphor.services*")
        .should_not_import("gaphor*")
        .should_not_import("gi*")
        .check(gaphor, skip_type_checking=True)
    )


def test_storage_package():
    (
        archrule("Storage only depends on core functionality")
        .match("gaphor.storage*")
        .exclude("*.tests.*")
        .may_import(*GAPHOR_CORE, *GLIB, *PANGO)
        .may_import("gaphor.diagram*")
        .may_import("gaphor.storage*")
        .may_import("gaphor.application", "gaphor.services.componentregistry")
        .should_not_import("gaphor*")
        .should_not_import("gi*")
        .check(gaphor, skip_type_checking=True)
    )


def test_modeling_languages_should_not_depend_on_ui_package():
    (
        archrule("Modeling languages should not depend on the UI package")
        .match("gaphor.C4Model*", "gaphor.RAAML*", "gaphor.SysML*", "gaphor.UML*")
        .may_import("gaphor.ui.filedialog")
        .may_import("gaphor.ui.errordialog")
        .should_not_import("gaphor.ui*")
        .check(gaphor)
    )


def test_moduling_languages_should_initialize_without_gtk():
    modeling_languages: list[str] = [
        c.__module__ for c in load_entry_points("gaphor.modelinglanguages").values()
    ]
    assert modeling_languages

    (
        archrule("No GTK dependency for modeling languages")
        .match(*modeling_languages)
        .should_not_import(*UI_LIBRARIES)
        .check(gaphor)
    )


def test_uml_package_does_not_depend_on_other_modeling_languages():
    (
        archrule("No modeling language dependencies for UML")
        .match("gaphor.UML*")
        .exclude("*.tests.*")
        .should_not_import("gaphor.C4Model*", "gaphor.RAAML*", "gaphor.SysML*")
        .check(gaphor, only_toplevel_imports=True)
    )


def test_allow_main_application_to_configure_i18n():
    (
        archrule(
            "The main application should configure i18n, so it should not be imported anywhere yet"
        )
        .match("gaphor.__main__")
        .match("gaphor.main")
        .should_not_import("gaphor.i18n")
        .check(gaphor, only_toplevel_imports=True)
    )



"""Test Stereotype Property Page."""

from gi.repository import Gtk

from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.profiles.stereotypepropertypages import StereotypePage


def test_stereotype_page_with_no_stereotype(element_factory, event_manager):
    subject = element_factory.create(UML.Class)

    editor = StereotypePage(subject, event_manager)
    page = editor.construct()

    assert page is None


def test_stereotype_page_with_stereotype(element_factory, event_manager):
    subject = element_factory.create(UML.Class)

    # Create a stereotype applicable to Class types:
    metaclass = element_factory.create(UML.Class)
    metaclass.name = "Class"
    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "NewStereotype"
    UML.recipes.create_extension(metaclass, stereotype)
    attr = element_factory.create(UML.Property)
    attr.name = "Property"

    editor = StereotypePage(subject, event_manager)
    page = editor.construct()

    stereotype_view = find(page, "stereotype-list")

    assert isinstance(stereotype_view, Gtk.ColumnView)
    assert len(stereotype_view.get_model()) == 1
    assert page is not None



"""Test classifier stereotypes attributes using component items."""

import pytest

from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.UML.classes.component import ComponentItem
from gaphor.UML.sanitizerservice import SanitizerService


def compartments(item):
    return item.shape.children[1:]


@pytest.fixture(autouse=True)
def sanitizer(event_manager):
    SanitizerService(event_manager)


@pytest.fixture
def case(element_factory):
    cls = element_factory.create(UML.Class)
    cls.name = "Component"
    st1_ = element_factory.create(UML.Stereotype)
    st1_.name = "st1"
    st2_ = element_factory.create(UML.Stereotype)
    st2_.name = "st2"

    attr = element_factory.create(UML.Property)
    attr.name = "st1_attr_1"
    st1_.ownedAttribute = attr
    attr = element_factory.create(UML.Property)
    attr.name = "st1_attr_2"
    st1_.ownedAttribute = attr

    attr = element_factory.create(UML.Property)
    attr.name = "st2_attr_1"
    st2_.ownedAttribute = attr

    class Case:
        st1 = st1_
        st2 = st2_
        ext1 = UML.recipes.create_extension(cls, st1_)
        ext2 = UML.recipes.create_extension(cls, st2_)

    return Case


def test_adding_slot(case, create):
    """Test if stereotype attribute information is added when slot is added."""
    c = create(ComponentItem, UML.Component)

    c.show_stereotypes = True
    instance_spec = UML.recipes.apply_stereotype(c.subject, case.st1)

    # test precondition
    assert not compartments(c)

    slot = UML.recipes.add_slot(instance_spec, case.st1.ownedAttribute[0])
    UML.recipes.set_slot_value(slot, "foo")

    assert len(compartments(c)) == 1


def test_removing_last_slot(case, create):
    """Test removing last slot."""
    c = create(ComponentItem, UML.Component)

    c.show_stereotypes = True
    instance_spec = UML.recipes.apply_stereotype(c.subject, case.st1)

    slot = UML.recipes.add_slot(instance_spec, case.st1.ownedAttribute[0])
    UML.recipes.set_slot_value(slot, "foo")

    # test precondition
    assert compartments(c)

    del instance_spec.slot[slot]
    assert not compartments(c)


def test_deleting_extension(case, create):
    """Test if stereotype is removed when extension is deleted."""
    c = create(ComponentItem, UML.Component)

    c.show_stereotypes = True

    instance_spec = UML.recipes.apply_stereotype(c.subject, case.st1)
    slot = UML.recipes.add_slot(instance_spec, case.st1.ownedAttribute[0])
    UML.recipes.set_slot_value(slot, "foo")

    # test precondition
    assert len(compartments(c)) == 1
    assert len(c.subject.appliedStereotype) == 1

    case.ext1.unlink()
    assert len(c.subject.appliedStereotype) == 0
    assert len(compartments(c)) == 0


def test_deleting_stereotype(case, create):
    """Test if stereotype is removed when stereotype is deleted."""
    c = create(ComponentItem, UML.Component)

    c.show_stereotypes = True

    st1 = case.st1
    instance_spec = UML.recipes.apply_stereotype(c.subject, st1)
    slot = UML.recipes.add_slot(instance_spec, case.st1.ownedAttribute[0])
    UML.recipes.set_slot_value(slot, "foo")

    # test precondition
    assert len(compartments(c)) == 1
    assert len(c.subject.appliedStereotype) == 1

    st1.unlink()
    assert len(c.subject.appliedStereotype) == 0
    assert len(compartments(c)) == 0


def test_removing_stereotype_attribute(case, element_factory, create):
    """Test if stereotype instance specification is destroyed when stereotype
    attribute is removed."""
    c = create(ComponentItem, UML.Component)

    c.show_stereotypes = True

    # test precondition
    assert len(compartments(c)) == 0
    obj = UML.recipes.apply_stereotype(c.subject, case.st1)
    # test precondition
    assert len(compartments(c)) == 0

    assert len(element_factory.lselect(UML.Slot)) == 0

    attr = case.st1.ownedAttribute[0]
    slot = UML.recipes.add_slot(obj, attr)
    UML.recipes.set_slot_value(slot, "foo")
    assert len(obj.slot) == 1
    assert len(element_factory.lselect(UML.Slot)) == 1
    assert slot.definingFeature

    assert compartments(c)

    attr.unlink()
    assert len(obj.slot) == 0
    assert 0 == len(element_factory.lselect(UML.Slot))
    assert not compartments(c)


def test_stereotype_attributes_status_saving(
    case, element_factory, create, saver, loader
):
    """Test stereotype attributes status saving."""
    c = create(ComponentItem, UML.Component)

    c.show_stereotypes = True
    UML.recipes.apply_stereotype(c.subject, case.st1)
    obj = UML.recipes.apply_stereotype(c.subject, case.st2)

    # change attribute of 2nd stereotype
    attr = case.st2.ownedAttribute[0]
    slot = UML.recipes.add_slot(obj, attr)
    UML.recipes.set_slot_value(slot, "st2 test21")

    data = saver()
    loader(data)
    diagram = next(element_factory.select(Diagram))

    item = next(diagram.select(ComponentItem))
    assert item.show_stereotypes
    assert len(compartments(c)) == 1


def test_saving_stereotype_attributes(case, element_factory, create, saver, loader):
    """Test stereotype attributes saving."""
    c = create(ComponentItem, UML.Component)

    c.show_stereotypes = True

    UML.recipes.apply_stereotype(c.subject, case.st1)
    UML.recipes.apply_stereotype(c.subject, case.st2)

    assert len(case.st1.ownedAttribute) == 3
    attr1, attr2, attr3 = case.st1.ownedAttribute
    assert attr1.name == "st1_attr_1", attr1.name
    assert attr2.name == "st1_attr_2", attr2.name
    assert attr3.name == "baseClass", attr3.name

    obj = c.subject.appliedStereotype[0]
    slot = UML.recipes.add_slot(obj, attr1)
    UML.recipes.set_slot_value(slot, "st1 test1")
    slot = UML.recipes.add_slot(obj, attr2)
    UML.recipes.set_slot_value(slot, "st1 test2")

    data = saver()
    loader(data)
    diagram = next(element_factory.select(Diagram))

    item = next(diagram.select(ComponentItem))
    el = item.subject
    assert len(el.appliedStereotype) == 2

    # check if stereotypes are properly applied
    names = sorted(obj.classifier[0].name for obj in el.appliedStereotype)
    assert ["st1", "st2"] == names

    # two attributes were changed for stereotype st1, so 2 slots
    obj = el.appliedStereotype[0]
    assert len(obj.slot) == 2
    assert "st1_attr_1" == obj.slot[0].definingFeature.name
    assert "st1 test1" == UML.recipes.get_slot_value(obj.slot[0])
    assert "st1_attr_2" == obj.slot[1].definingFeature.name
    assert "st1 test2" == UML.recipes.get_slot_value(obj.slot[1])

    # no stereotype st2 attribute changes, no slots
    obj = el.appliedStereotype[1]
    assert len(obj.slot) == 0



from gaphor.babel import extract_gaphor


def test_babel(test_models):
    with (test_models / "test-model.gaphor").open(encoding="utf-8") as model:
        texts = list(extract_gaphor(model, (), (), None))

    assert (None, "gettext", "Element", []) in texts



import pytest

from gaphor.ui.help import new_builder


@pytest.fixture(autouse=True)
def builder():
    return new_builder("preferences")


def test_preferences_dark_mode(builder):
    style_variant = builder.get_object("style_variant")
    assert style_variant.get_selected() == 0


def test_preferences_use_english(builder):
    use_english = builder.get_object("use_english")
    assert use_english.get_active() == 0



import pytest

from gaphor import UML
from gaphor.i18n import gettext
from gaphor.UML.treemodel import (
    Branch,
    RelationshipItem,
    Root,
    TreeItem,
    TreeModel,
    tree_item_sort,
)


@pytest.fixture
def tree_model(event_manager, element_factory):
    return TreeModel(event_manager, element_factory)


class ItemChangedHandler:
    def __init__(self):
        self.added = 0
        self.removed = 0
        self.positions = []

    def __call__(self, _obj, position, removed, added):
        self.positions.append(position)
        self.removed += removed
        self.added += added


def test_tree_item_gtype():
    assert TreeItem.__gtype__.name == "gaphor+UML+treemodel+TreeItem"


def test_branch_add_element(element_factory):
    branch = Branch()
    element = element_factory.create(UML.Class)

    branch.append(element)

    assert len(branch) == 1
    assert branch[0].element is element


def test_branch_remove_element(element_factory):
    branch = Branch()
    element = element_factory.create(UML.Class)
    branch.append(element)

    branch.remove(element)

    assert len(branch) == 0


def test_branch_add_relationship(element_factory):
    branch = Branch()
    element = element_factory.create(UML.Association)

    branch.append(element)

    assert len(branch) == 1
    assert isinstance(branch[0], RelationshipItem)
    assert branch.relationships[0].element is element


def test_tree_model_add_element(tree_model, element_factory):
    element = element_factory.create(UML.Class)

    tree_model.add_element(element)
    tree_item = tree_model.tree_item_for_element(element)

    assert tree_item.element is element


def test_tree_model_add_nested_element(tree_model, element_factory):
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    class_.package = package
    tree_model.add_element(package)
    tree_model.add_element(class_)
    tree_model.child_model(tree_model.tree_item_for_element(package))
    class_item = tree_model.tree_item_for_element(class_)
    package_item = tree_model.tree_item_for_element(package)

    assert tree_model.branches.get(package_item) is not None
    assert tree_model.branches.get(package_item)[0] is class_item
    assert tree_model.branches.get(class_item) is None


def test_tree_model_add_nested_element_in_reverse_order(tree_model, element_factory):
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    class_.package = package
    tree_model.add_element(class_)
    tree_model.add_element(package)
    tree_model.child_model(tree_model.tree_item_for_element(package))
    class_item = tree_model.tree_item_for_element(class_)
    package_item = tree_model.tree_item_for_element(package)

    assert tree_model.branches.get(package_item) is not None
    assert tree_model.branches.get(package_item)[0] is class_item
    assert tree_model.branches.get(class_item) is None


def test_tree_model_remove_element(tree_model, element_factory):
    element = element_factory.create(UML.Class)
    tree_model.add_element(element)

    tree_model.remove_element(element)
    tree_item = tree_model.tree_item_for_element(element)

    assert tree_item is None


def test_tree_model_remove_nested_element(tree_model, element_factory):
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    class_.package = package
    tree_model.add_element(package)
    tree_model.add_element(class_)
    tree_model.child_model(tree_model.tree_item_for_element(package))
    tree_model.remove_element(class_)

    assert len(tree_model.branches) == 1
    assert Root in tree_model.branches
    assert tree_model.tree_item_for_element(package) is not None
    assert tree_model.tree_item_for_element(class_) is None


def test_tree_model_remove_package_with_nested_element(tree_model, element_factory):
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    class_.package = package
    tree_model.add_element(package)
    tree_model.add_element(class_)
    tree_model.remove_element(package)

    assert tree_model.tree_item_for_element(package) is None
    assert len(tree_model.branches) == 1
    assert Root in tree_model.branches
    assert tree_model.tree_item_for_element(package) is None
    assert tree_model.tree_item_for_element(class_) is None


def test_tree_model_remove_from_different_owner(tree_model, element_factory):
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    tree_model.add_element(package)
    tree_model.add_element(class_)
    class_.package = package
    tree_model.remove_element(class_)

    assert tree_model.tree_item_for_element(package) is not None
    assert len(tree_model.root) == 1


def test_tree_model_change_owner(tree_model, element_factory):
    class_ = element_factory.create(UML.Class)
    package = element_factory.create(UML.Package)

    tree_model.add_element(package)
    tree_model.add_element(class_)
    class_.package = package
    tree_model.remove_element(class_)
    tree_model.add_element(class_)
    package_item = tree_model.tree_item_for_element(package)
    tree_model.child_model(package_item)
    class_item = tree_model.tree_item_for_element(class_)
    package_model = tree_model.branches[package_item]

    assert package_item in tree_model.root
    assert class_item not in tree_model.root
    assert class_item in package_model


def test_tree_model_relationship_subtree(tree_model, element_factory):
    package = element_factory.create(UML.Package)
    association = element_factory.create(UML.Association)
    association.package = package

    tree_model.add_element(package)
    tree_model.add_element(association)
    package_item = tree_model.tree_item_for_element(package)
    tree_model.child_model(package_item)
    association_item = tree_model.tree_item_for_element(association)
    package_model = tree_model.branches[package_item]
    relationship_item = package_model[0]
    branch = tree_model.owner_branch_for_element(association)
    relationship_model = branch.relationships

    assert package_model
    assert isinstance(relationship_item, RelationshipItem)
    assert relationship_item.readonly_text == gettext("<Relationships>")
    assert association_item is relationship_model.get_item(0)


def test_tree_model_second_relationship(tree_model, element_factory):
    package = element_factory.create(UML.Package)
    association = element_factory.create(UML.Association)
    association.package = package

    tree_model.add_element(package)
    tree_model.add_element(association)
    package_item = tree_model.tree_item_for_element(package)
    tree_model.child_model(package_item)
    branch = tree_model.owner_branch_for_element(association)
    relationship_model = branch.relationships

    new_association = element_factory.create(UML.Association)
    new_association.package = package
    tree_model.add_element(new_association)

    association_item = tree_model.tree_item_for_element(association)
    new_association_item = tree_model.tree_item_for_element(new_association)

    assert association_item
    assert new_association_item
    assert association_item is relationship_model[0]
    assert new_association_item is relationship_model[1]


def test_tree_model_remove_relationship(tree_model, element_factory):
    package = element_factory.create(UML.Package)
    association = element_factory.create(UML.Association)
    association.package = package
    tree_model.add_element(package)
    tree_model.add_element(association)
    package_item = tree_model.tree_item_for_element(package)
    tree_model.child_model(package_item)

    tree_model.remove_element(association)
    package_model = tree_model.branches.get(package_item)

    assert not package_model


def test_tree_model_sort_relationship_item_first():
    a = TreeItem(UML.Package())
    a.readonly_text = "a"
    b = TreeItem(UML.Package())
    b.readonly_text = "b"
    r = RelationshipItem(None)

    assert tree_item_sort(r, b) == -1
    assert tree_item_sort(a, r) == 1
    assert tree_item_sort(a, b) == -1
    assert tree_item_sort(b, a) == 1


def test_element_with_member_and_no_owner(tree_model, element_factory):
    property = element_factory.create(UML.Property)
    association = element_factory.create(UML.Association)
    association.memberEnd = property
    tree_model.add_element(association)
    tree_model.add_element(property)
    association_item = tree_model.tree_item_for_element(association)
    tree_model.child_model(association_item)
    property_item = tree_model.tree_item_for_element(property)

    association_model = tree_model.branches.get(association_item)

    assert property_item in association_model.elements



from gaphor import UML
from gaphor.diagram.drop import drop
from gaphor.UML.interactions import LifelineItem


def test_drop_class_on_lifeline(element_factory, create, diagram):
    lifeline_item = create(LifelineItem, UML.Lifeline)
    interface = element_factory.create(UML.Interface)

    drop(interface, diagram, 10, 10)

    assert lifeline_item.subject.represents is interface



from uuid import uuid1

import pytest

from gaphor import UML
from gaphor.core.changeset.compare import set_value_change_property_value
from gaphor.core.modeling import (
    Diagram,
    ElementChange,
    RefChange,
    ValueChange,
)
from gaphor.ui.modelmerge.organize import organize_changes
from gaphor.UML.diagramitems import ClassItem


@pytest.fixture(autouse=True)
def mock_gettext(monkeypatch):
    monkeypatch.setattr("gaphor.ui.modelmerge.organize.gettext", lambda s: s)


@pytest.fixture
def change(element_factory):
    def _change(change_type, **kwargs):
        assert "op" in kwargs
        if "element_id" not in kwargs:
            kwargs["element_id"] = str(uuid1())

        if change_type is ElementChange:
            assert "element_name" in kwargs
        elif change_type is RefChange:
            assert "property_name" in kwargs
            if "property_ref" not in kwargs:
                kwargs["property_ref"] = str(uuid1())
        elif change_type is ValueChange:
            assert "property_name" in kwargs

        c = element_factory.create(change_type)
        for n, v in kwargs.items():
            setattr(c, n, v)

        return c

    return _change


def test_add_element(element_factory, modeling_language):
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "1234"
    change.element_name = "Diagram"

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Add element of type Diagram"
    assert change in add_element.elements


def test_remove_element(element_factory, modeling_language):
    diagram = element_factory.create(Diagram)
    diagram.name = "my diagram"
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "remove"
    change.element_id = diagram.id
    change.element_name = "Diagram"

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Remove element “my diagram”"
    assert change in add_element.elements


def test_add_element_with_attribute_update(element_factory, modeling_language):
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "1234"
    change.element_name = "Diagram"
    vchange: ValueChange = element_factory.create(ValueChange)
    vchange.op = "update"
    vchange.element_id = "1234"
    vchange.property_name = "name"
    set_value_change_property_value(vchange, "my diagram")

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Add Diagram “my diagram”"
    assert change in add_element.elements
    assert vchange in add_element.elements
    assert not add_element.children


def test_remove_element_with_attribute_update(
    element_factory, modeling_language, change
):
    diagram = element_factory.create(Diagram)
    diagram.name = "my diagram"
    echange: ElementChange = change(
        ElementChange, element_name="Diagram", op="remove", element_id=diagram.id
    )
    vchange: ValueChange = change(
        ValueChange,
        op="update",
        element_id=diagram.id,
        property_name="name",
    )
    set_value_change_property_value(vchange, "my diagram")

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Remove element “my diagram”"
    assert echange in add_element.elements
    assert vchange in add_element.elements
    assert not add_element.children


def test_update_diagram_attribute(element_factory, modeling_language, change):
    diagram = element_factory.create(Diagram)
    diagram.name = "my diagram"
    vchange: ValueChange = change(
        ValueChange,
        op="update",
        element_id=diagram.id,
        property_name="name",
    )
    set_value_change_property_value(vchange, "my diagram")

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Update diagram “my diagram”"
    assert vchange in add_element.elements
    assert not add_element.children


def test_update_model_attribute(element_factory, modeling_language, change):
    klass = element_factory.create(UML.Class)
    vchange: ValueChange = change(
        ValueChange,
        op="update",
        element_id=klass.id,
        property_name="name",
    )
    set_value_change_property_value(vchange, "my class")

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Update element of type “Class”"
    assert vchange in add_element.elements
    assert not add_element.children


def test_update_reference_without_name(element_factory, modeling_language):
    diagram = element_factory.create(Diagram)
    diagram.name = "my diagram"
    element = element_factory.create(UML.Element)
    vchange: RefChange = element_factory.create(RefChange)
    vchange.op = "add"
    vchange.element_id = diagram.id
    vchange.property_name = "element"
    vchange.property_ref = element.id

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Update diagram “my diagram”"
    assert not add_element.elements
    assert add_element.children
    assert vchange in add_element.children[0].elements
    assert add_element.children[0].label == "Add relation “element”"


def test_update_reference_with_name(element_factory, modeling_language):
    diagram = element_factory.create(Diagram)
    diagram.name = "my diagram"
    element = element_factory.create(UML.Element)
    vchange: RefChange = element_factory.create(RefChange)
    vchange.op = "add"
    vchange.element_id = element.id
    vchange.property_name = "ownedDiagram"
    vchange.property_ref = diagram.id

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Update element of type “Element”"
    assert not add_element.elements
    assert add_element.children
    assert vchange in add_element.children[0].elements
    assert (
        add_element.children[0].label == "Add relation “ownedDiagram” to “my diagram”"
    )


def test_remove_reference_with_name(element_factory, modeling_language):
    diagram = element_factory.create(Diagram)
    diagram.name = "my diagram"
    element = element_factory.create(UML.Element)
    vchange: RefChange = element_factory.create(RefChange)
    vchange.op = "remove"
    vchange.element_id = element.id
    vchange.property_name = "ownedDiagram"
    vchange.property_ref = diagram.id

    tree = list(organize_changes(element_factory, modeling_language))

    add_element = tree[0]

    assert len(tree) == 1
    assert add_element.label == "Update element of type “Element”"
    assert not add_element.elements
    assert add_element.children
    assert vchange in add_element.children[0].elements
    assert (
        add_element.children[0].label
        == "Remove relation “ownedDiagram” to “my diagram”"
    )


def test_add_diagram_with_reference(element_factory, modeling_language, change):
    add_diagram = change(ElementChange, op="add", element_name="Diagram")
    ref = change(
        RefChange,
        op="add",
        element_id=add_diagram.element_id,
        property_name="element",
    )

    tree = list(organize_changes(element_factory, modeling_language))

    assert add_diagram in tree[0].elements
    assert tree[0].children
    assert ref in tree[0].children[0].elements


def test_unresolvable_reference_element_id(element_factory, modeling_language, change):
    change(
        RefChange,
        op="add",
        element_id="12345",
        property_name="element",
    )

    tree = list(organize_changes(element_factory, modeling_language))

    assert not tree


def test_unresolvable_reference_property_ref(
    element_factory, modeling_language, change
):
    change(
        RefChange,
        op="add",
        property_name="element",
        property_ref="12345",
    )

    tree = list(organize_changes(element_factory, modeling_language))

    assert not tree


def test_add_diagram_contains_presentation(element_factory, modeling_language, change):
    add_diagram = change(ElementChange, op="add", element_name="Diagram")
    add_class_item = change(ElementChange, op="add", element_name="ClassItem")
    change(
        RefChange,
        op="add",
        element_id=add_diagram.element_id,
        property_name="ownedPresentation",
        property_ref=add_class_item.element_id,
    )

    tree = list(organize_changes(element_factory, modeling_language))

    assert add_diagram in tree[0].elements
    assert tree[0].children
    assert add_class_item in tree[0].children[0].elements


def test_add_diagram_contains_presentation_with_subject(
    element_factory, modeling_language, change
):
    add_diagram = change(ElementChange, op="add", element_name="Diagram")
    add_class_item = change(ElementChange, op="add", element_name="ClassItem")
    change(
        RefChange,
        op="add",
        element_id=add_diagram.element_id,
        property_name="ownedPresentation",
        property_ref=add_class_item.element_id,
    )
    add_class = change(ElementChange, op="add", element_name="Class")
    change(
        RefChange,
        op="add",
        element_id=add_class_item.element_id,
        property_name="subject",
        property_ref=add_class.element_id,
    )

    tree = list(organize_changes(element_factory, modeling_language))

    assert add_diagram in tree[0].elements
    assert tree[0].children
    assert add_class_item in tree[0].children[0].elements
    assert tree[0].children[0].children
    assert add_class in tree[0].children[0].children[0].elements


def test_add_presentation_to_existing_diagram(
    element_factory, modeling_language, change
):
    diagram = element_factory.create(Diagram)
    add_class_item = change(ElementChange, op="add", element_name="ClassItem")
    change(
        RefChange,
        op="add",
        element_id=diagram.id,
        property_name="ownedPresentation",
        property_ref=add_class_item.element_id,
    )

    tree = list(organize_changes(element_factory, modeling_language))

    assert not tree[0].elements
    assert tree[0].children
    assert add_class_item in tree[0].children[0].elements


def test_update_presentation_to_existing_diagram(
    element_factory, modeling_language, change
):
    diagram = element_factory.create(Diagram)
    diagram.name = "my diagram"
    class_item = diagram.create(ClassItem)
    vchange = change(
        ValueChange,
        op="update",
        element_id=class_item.id,
        property_name="name",
    )
    set_value_change_property_value(vchange, "new name")

    tree = list(organize_changes(element_factory, modeling_language))

    assert tree[0].label == "Update element of type “ClassItem”"
    assert tree[0].elements
    assert not tree[0].children
    assert vchange in tree[0].elements


def test_add_diagram_contains_presentation_with_subject_and_composite_element(
    element_factory, modeling_language, change
):
    add_diagram = change(ElementChange, op="add", element_name="Diagram")
    add_class_item = change(ElementChange, op="add", element_name="ClassItem")
    change(
        RefChange,
        op="add",
        element_id=add_diagram.element_id,
        property_name="ownedPresentation",
        property_ref=add_class_item.element_id,
    )
    add_class = change(ElementChange, op="add", element_name="Class")
    change(
        RefChange,
        op="add",
        element_id=add_class_item.element_id,
        property_name="subject",
        property_ref=add_class.element_id,
    )
    add_property = change(ElementChange, op="add", element_name="Property")
    change(
        RefChange,
        op="add",
        element_id=add_class.element_id,
        property_name="ownedAttribute",
        property_ref=add_property.element_id,
    )

    tree = list(organize_changes(element_factory, modeling_language))

    assert len(tree) == 1
    assert add_diagram in tree[0].elements
    assert tree[0].children
    assert add_class_item in tree[0].children[0].elements
    assert tree[0].children[0].children
    assert add_class in tree[0].children[0].children[0].elements
    assert tree[0].children[0].children[0].children
    assert add_property in tree[0].children[0].children[0].children[0].elements



import gaphas
import pytest

from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling import ElementFactory, Presentation, StyleSheet
from gaphor.core.modeling.diagram import Diagram, StyledDiagram, StyledItem


@pytest.fixture
def event_manager():
    return EventManager()


@pytest.fixture
def element_factory(event_manager):
    return ElementFactory(event_manager)


@pytest.fixture
def diagram(element_factory):
    diagram = element_factory.create(Diagram)
    yield diagram
    diagram.unlink()


class DemoItem(gaphas.Element, Presentation):
    def __init__(self, diagram, id):
        super().__init__(connections=diagram.connections, diagram=diagram, id=id)


def test_name_does_not_have_item_suffix(diagram: Diagram):
    item = diagram.create(DemoItem)
    node = StyledItem(item)

    assert node.name() == "demo"


def test_diagram_is_parent_of_item(diagram: Diagram):
    item = diagram.create(DemoItem)
    node = StyledItem(item)

    assert node.parent().name() == "diagram"


def test_diagram_has_no_parent(diagram):
    node = StyledDiagram(diagram)

    assert node.parent() is None


def test_style_sheet_has_default_style():
    style_sheet = StyleSheet()

    assert "diagram {" in style_sheet.styleSheet



from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.classes.interface import InterfaceItem
from gaphor.UML.classes.interfacerealization import InterfaceRealizationItem
from gaphor.UML.classes.klass import ClassItem


def test_interface_realization_multiple_connection(create, element_factory):
    c1 = create(InterfaceItem, UML.Interface)
    c2 = create(ClassItem, UML.Class)
    rel1 = create(InterfaceRealizationItem)

    connect(rel1, rel1.head, c1)
    connect(rel1, rel1.tail, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    rel2 = diagram2.create(InterfaceRealizationItem)

    connect(rel2, rel2.head, c3)
    connect(rel2, rel2.tail, c4)

    assert rel1.subject is rel2.subject


def test_interface_realization_reconnect_in_new_diagram(create, element_factory):
    c1 = create(InterfaceItem, UML.Interface)
    c2 = create(ClassItem, UML.Class)
    rel = create(InterfaceRealizationItem)

    connect(rel, rel.head, c1)
    connect(rel, rel.tail, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(InterfaceItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    rel2 = diagram2.create(InterfaceRealizationItem)

    connect(rel2, rel2.head, c3)
    connect(rel2, rel2.tail, c4)
    assert rel.subject is rel2.subject

    c5 = diagram2.create(InterfaceItem, subject=element_factory.create(UML.Interface))
    connect(rel2, rel2.head, c5)

    assert rel.subject is not rel2.subject
    assert c1.subject in rel.subject.supplier
    assert c2.subject in rel.subject.client
    assert c5.subject in rel2.subject.supplier
    assert c4.subject in rel2.subject.client


def test_interface_realization_reconnect_twice_in_new_diagram(create, element_factory):
    c1 = create(InterfaceItem, UML.Interface)
    c2 = create(ClassItem, UML.Class)
    rel = create(InterfaceRealizationItem)

    connect(rel, rel.head, c1)
    connect(rel, rel.tail, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(InterfaceItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    rel2 = diagram2.create(InterfaceRealizationItem)

    connect(rel2, rel2.head, c3)
    connect(rel2, rel2.tail, c4)
    assert rel.subject is rel2.subject

    c5 = diagram2.create(InterfaceItem, subject=element_factory.create(UML.Interface))
    connect(rel2, rel2.head, c5)
    connect(rel2, rel2.head, c3)

    assert rel.subject is rel2.subject
    assert c1.subject in rel.subject.supplier
    assert c2.subject in rel.subject.client



import pytest

from gaphor.core.modeling.modelinglanguage import (
    CoreModelingLanguage,
    MockModelingLanguage,
)
from gaphor.SysML.modelinglanguage import SysMLModelingLanguage
from gaphor.UML.modelinglanguage import UMLModelingLanguage


@pytest.fixture
def modeling_language():
    return MockModelingLanguage(
        CoreModelingLanguage(), UMLModelingLanguage(), SysMLModelingLanguage()
    )



"""Unit tests for simple items."""

from gaphor.diagram.general.simpleitem import Box, Ellipse, Line, LineEndStyle


def test_line(diagram):
    line = diagram.create(Line)

    assert line
    assert line.head_end == LineEndStyle.none
    assert line.tail_end == LineEndStyle.none


def test_box(diagram):
    assert diagram.create(Box)


def test_ellipse(diagram):
    assert diagram.create(Ellipse)



import os

# https://github.com/pyinstaller/pyinstaller/issues/6100
# On one Windows computer, PyInstaller was adding a ; to
# end of the path, this removes it if it exists
if os.environ["PATH"][-1] == ";":
    os.environ["PATH"] = os.environ["PATH"][:-1]

# Check for and remove two semicolons in path
os.environ["PATH"] = os.environ["PATH"].replace(";;", ";")



from gaphor import UML
from gaphor.diagram.copypaste import copy_full, paste_full
from gaphor.diagram.tests.test_copypaste_link import two_classes_and_a_generalization
from gaphor.UML.classes import PackageItem


def test_copy_connected_item(diagram, element_factory):
    spc_cls_item, gen_cls_item, gen_item = two_classes_and_a_generalization(
        diagram, element_factory
    )

    buffer = copy_full({spc_cls_item})
    (new_cls_item,) = paste_full(buffer, diagram)

    assert new_cls_item.subject
    assert gen_item.subject.general is gen_cls_item.subject
    assert gen_item.subject.specific is spc_cls_item.subject


def test_full_copy_package_from_owned_diagram(element_factory):
    diagram: UML.Diagram = element_factory.create(UML.Diagram)
    package_item = diagram.create(
        PackageItem, subject=element_factory.create(UML.Package)
    )
    package_item.subject.ownedDiagram = diagram

    copy_buffer = copy_full([package_item], element_factory.lookup)

    new_diagram: UML.Diagram = element_factory.create(UML.Diagram)

    (new_package_item,) = paste_full(copy_buffer, new_diagram)

    assert new_package_item.subject
    assert new_package_item.subject is not package_item.subject


def test_full_copy_package_from_owned_diagram_in_super_package(element_factory):
    diagram: UML.Diagram = element_factory.create(UML.Diagram)
    package_item = diagram.create(
        PackageItem, subject=element_factory.create(UML.Package)
    )
    package_item.subject.ownedDiagram = diagram

    copy_buffer = copy_full([package_item], element_factory.lookup)

    new_diagram: UML.Diagram = element_factory.create(UML.Diagram)
    new_diagram.element = element_factory.create(UML.Package)

    (new_package_item,) = paste_full(copy_buffer, new_diagram)

    assert diagram.owner is package_item.subject
    assert new_package_item.subject
    assert new_package_item.subject is not package_item.subject
    assert new_package_item.subject.owner is new_diagram.owner


def test_copy_slot(diagram, element_factory):
    slot = element_factory.create(UML.Slot)
    slot.value = element_factory.create(UML.LiteralInteger)
    slot.value.value = 1

    buffer = copy_full({slot})
    paste_full(buffer, diagram)

    new_slot = next(p for p in element_factory.select(UML.Slot) if p is not slot)

    assert new_slot.value
    assert new_slot.value.value == 1

    assert new_slot.value is not slot.value



import pytest

import gaphor.UML.uml as UML
from gaphor.diagram.group import can_group, group, owner, ungroup
from gaphor.UML.recipes import create_association, set_navigability


def test_can_group_package_and_class(element_factory):
    package = element_factory.create(UML.Package)
    klass = element_factory.create(UML.Class)

    assert can_group(package, klass)


def test_group_package_and_class(element_factory):
    package = element_factory.create(UML.Package)
    klass = element_factory.create(UML.Class)

    assert group(package, klass)

    assert klass.owner is package


def test_ungroup(element_factory):
    package = element_factory.create(UML.Package)
    klass = element_factory.create(UML.Class)
    group(package, klass)

    assert ungroup(package, klass)

    assert klass.package is None


def test_do_not_ungroup_wrong_parent(element_factory):
    package = element_factory.create(UML.Package)
    klass = element_factory.create(UML.Class)
    wrong_package = element_factory.create(UML.Package)
    group(package, klass)

    assert not ungroup(wrong_package, klass)

    assert klass.package is package


def test_group_package_and_package(element_factory):
    package = element_factory.create(UML.Package)
    parent = element_factory.create(UML.Package)

    assert group(parent, package)

    assert package.owner is parent


def test_ungroup_package(element_factory):
    package = element_factory.create(UML.Package)
    parent = element_factory.create(UML.Package)
    group(parent, package)

    assert ungroup(parent, package)

    assert package.nestingPackage is None


def test_do_not_ungroup_package_wrong_parent(element_factory):
    package = element_factory.create(UML.Package)
    parent = element_factory.create(UML.Package)
    wrong_parent = element_factory.create(UML.Package)
    group(parent, package)

    assert not ungroup(wrong_parent, package)

    assert package.nestingPackage is parent


def test_group_class_and_activity(element_factory):
    klass = element_factory.create(UML.Class)
    activity = element_factory.create(UML.Activity)

    assert group(klass, activity)

    assert activity.owner is klass


def test_group_class_and_activity_from_package(element_factory):
    klass = element_factory.create(UML.Class)
    activity = element_factory.create(UML.Activity)

    assert group(klass, activity)

    assert activity.owner is klass


def test_ungroup_class_and_activity(element_factory):
    package = element_factory.create(UML.Package)
    klass = element_factory.create(UML.Class)
    activity = element_factory.create(UML.Activity)

    group(package, activity)
    group(klass, activity)

    assert activity.owner is klass


@pytest.mark.parametrize(
    "classifier_type,feature_type",
    [
        (UML.Artifact, UML.Operation),
        (UML.Class, UML.Operation),
        (UML.DataType, UML.Operation),
        (UML.Interface, UML.Operation),
        (UML.Artifact, UML.Property),
        (UML.Class, UML.Property),
        (UML.DataType, UML.Property),
        (UML.Interface, UML.Property),
    ],
)
def test_group_classifier_and_feature(classifier_type, feature_type, element_factory):
    classifier = element_factory.create(classifier_type)
    feature = element_factory.create(feature_type)
    classifier2 = element_factory.create(classifier_type)

    assert group(classifier, feature)
    assert group(classifier2, feature)

    assert feature.owner is classifier2


def test_should_not_group_association_end(element_factory):
    head_class = element_factory.create(UML.Class)
    tail_class = element_factory.create(UML.Class)
    association = create_association(head_class, tail_class)
    set_navigability(association, association.memberEnd[0], True)
    other_class = element_factory.create(UML.Class)
    end = association.memberEnd[0]

    assert not group(other_class, end)
    assert not ungroup(end.owner, end)


@pytest.mark.parametrize(
    "element_type",
    [
        UML.Slot,
        UML.Comment,
        UML.Image,
        UML.InstanceSpecification,
        UML.OccurrenceSpecification,
        UML.LiteralSpecification,
        UML.LiteralUnlimitedNatural,
        # The following elements are not shown if they have no owner
        UML.ConnectorEnd,
        UML.Parameter,
        UML.Pin,
        UML.Property,
        UML.StructuralFeature,
    ],
)
def test_element_with_no_owner(element_type):
    assert owner(element_type()) is None



from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.actions.partitionpage import PartitionPropertyPage


def test_partition_page(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.actions.PartitionItem, subject=element_factory.create(UML.ActivityPartition)
    )
    property_page = PartitionPropertyPage(item, event_manager)

    widget = property_page.construct()

    num_partitions = find(widget, "num-partitions")
    num_partitions.set_value(4)

    assert len(item.partition) == 4


def test_partition_name(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.actions.PartitionItem, subject=element_factory.create(UML.ActivityPartition)
    )
    property_page = PartitionPropertyPage(item, event_manager)
    property_page.construct()
    property_page.update_partitions(1)

    widget = property_page.construct_partition(item.partition[0])

    name = find(widget, "partition-name")
    name.set_text("Name")

    assert item.partition[0].name == "Name"


def test_partition_type(diagram, element_factory, event_manager):
    class_ = element_factory.create(UML.Class)
    class_.name = "Foobar"
    item = diagram.create(
        UML.actions.PartitionItem, subject=element_factory.create(UML.ActivityPartition)
    )
    property_page = PartitionPropertyPage(item, event_manager)
    property_page.construct()
    property_page.update_partitions(1)

    widget = property_page.construct_partition(item.partition[0])

    dropdown = find(widget, "partition-type")
    class_index = next(
        n for n, lv in enumerate(dropdown.get_model()) if lv.value == class_.id
    )
    dropdown.set_selected(class_index)

    assert item.partition[0].represents is class_



import re
import sys
from pathlib import Path

import babel.messages.pofile as pofile


def invalid(messages):
    placeholders = re.compile(r"{\w*}")
    html_entity = re.compile(r"&\w+;")

    for message in messages:
        if not message.string or message.fuzzy:
            continue

        if "%" in message.id:
            yield message, f"Old %-syntax used '{message.id}'"

        id_formats = placeholders.findall(message.id)
        str_formats = placeholders.findall(message.string)

        if "{}" in id_formats:
            yield message, f"Empty placeholder for '{message.id}': '{message.string}'"

        if set(id_formats) != set(str_formats):
            yield (
                message,
                f"Invalid placeholders for '{message.id}': '{message.string}'",
            )

        if html_entity.findall(message.string):
            yield message, f"Translation contains HTML entities: '{message.string}'"


def check_po_files():
    po_path: Path = Path(__file__).resolve().parent.parent / "po"
    have_errors = False
    for path in (path for path in po_path.iterdir() if path.suffix == ".po"):
        with path.open(encoding="utf-8") as po:
            messages = pofile.read_po(po)

        for message, error in invalid(messages):
            have_errors = True
            print(f"{path.name}:{message.lineno}: {error}.")  # noqa: T201
    return have_errors


if __name__ == "__main__":
    sys.exit(check_po_files())



"""Test Metaclass Property Page."""

import pytest
from gi.repository import Gtk

from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.profiles.metaclasspropertypage import MetaclassPropertyPage


@pytest.fixture
def class_(element_factory):
    class_ = element_factory.create(UML.Class)
    class_.name = "Class"
    yield class_
    del class_


def test_name_input_field_for_normal_class(class_, event_manager):
    """Test Metaclass Property Page not create for normal Class."""

    editor = MetaclassPropertyPage(class_, event_manager)
    page = editor.construct()

    assert not page


def test_name_selection_for_metaclass(element_factory, class_, event_manager):
    """Test the creation of Metaclass Property Page."""

    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "NewStereotype"
    UML.recipes.create_extension(class_, stereotype)

    editor = MetaclassPropertyPage(class_, event_manager)
    page = editor.construct()

    assert page

    dropdown = find(page, "metaclass-dropdown")

    assert type(dropdown) is Gtk.DropDown
    assert dropdown.get_selected_item().get_string() == "Class"

    class_.name = "Action"

    assert dropdown.get_selected_item().get_string() == "Action"



"""Test transition item and state vertices connections."""

import pytest

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect, get_connected
from gaphor.UML.states.finalstate import FinalStateItem
from gaphor.UML.states.pseudostates import PseudostateItem
from gaphor.UML.states.state import StateItem
from gaphor.UML.states.transition import TransitionItem


@pytest.fixture
def select(element_factory):
    return element_factory.lselect


def test_vertex_connect(create, select):
    v1 = create(StateItem, UML.State)
    v2 = create(StateItem, UML.State)

    t = create(TransitionItem)

    # connect vertices with transition
    connect(t, t.head, v1)
    connect(t, t.tail, v2)

    assert t.subject is not None

    assert len(select(UML.Transition)) == 1

    assert t.subject == v1.subject.outgoing[0]
    assert t.subject == v2.subject.incoming[0]
    assert t.subject.source == v1.subject
    assert t.subject.target == v2.subject


def test_vertex_reconnect(create, select, sanitizer_service):
    v1 = create(StateItem, UML.State)
    v2 = create(StateItem, UML.State)
    v3 = create(StateItem, UML.State)

    t = create(TransitionItem)

    # connect: v1 -> v2
    connect(t, t.head, v1)
    connect(t, t.tail, v2)

    s = t.subject
    s.name = "tname"
    specification = t.model.create(UML.LiteralString)
    specification.value = "tguard"
    specification.owningConstraint = s.guard
    s.guard.specification = specification

    # reconnect: v1 -> v3
    connect(t, t.tail, v3)

    assert s is not t.subject
    assert len(select(UML.Transition)) == 1

    assert t.subject == v1.subject.outgoing[0]
    assert t.subject == v3.subject.incoming[0]
    assert t.subject.source == v1.subject
    assert t.subject.target == v3.subject

    assert len(v2.subject.incoming) == 0
    assert len(v2.subject.outgoing) == 0


def test_vertex_disconnect(create, select):
    t = create(TransitionItem)
    v1 = create(StateItem, UML.State)
    v2 = create(StateItem, UML.State)

    connect(t, t.head, v1)
    connect(t, t.tail, v2)
    assert t.subject is not None

    assert len(select(UML.Transition)) == 1

    # test preconditions
    assert t.subject == v1.subject.outgoing[0]
    assert t.subject == v2.subject.incoming[0]

    disconnect(t, t.tail)
    assert t.subject is None

    disconnect(t, t.head)
    assert t.subject is None


def test_state_connect_to_same_item(create):
    """Test transition to state vertex connection."""
    v1 = create(StateItem, UML.State)
    t = create(TransitionItem)

    connect(t, t.head, v1)

    assert allow(t, t.tail, v1)


def test_initial_pseudostate_connect(create, select):
    v1 = create(PseudostateItem, UML.Pseudostate)
    v2 = create(StateItem, UML.State)

    t = create(TransitionItem)

    # connect head of transition to an initial pseudostate
    connect(t, t.head, v1)
    assert t.subject is None

    # connect tail of transition to a state
    connect(t, t.tail, v2)
    assert t.subject is not None

    assert len(select(UML.Transition)) == 1
    assert t.subject == v1.subject.outgoing[0]
    assert t.subject == v2.subject.incoming[0]


def test_initial_pseudostate_disconnect(create):
    """Test transition and initial pseudostate disconnection."""
    v1 = create(PseudostateItem, UML.Pseudostate)
    create(StateItem, UML.State)

    t = create(TransitionItem)
    assert t.subject is None

    # connect head of transition to an initial pseudostate
    connect(t, t.head, v1)
    assert get_connected(t, t.head)

    # perform the test
    disconnect(t, t.head)
    assert not get_connected(t, t.head)


def test_final_state_connect(create, select):
    """Test transition to final state connection."""
    v1 = create(StateItem, UML.State)
    v2 = create(FinalStateItem, UML.FinalState)
    t = create(TransitionItem)

    # connect head of transition to a state
    connect(t, t.head, v1)

    # check if transition can connect to final state
    glued = allow(t, t.tail, v2)
    assert glued
    # and connect tail of transition to final state
    connect(t, t.tail, v2)
    assert t.subject is not None

    assert len(select(UML.Transition)) == 1

    assert t.subject == v1.subject.outgoing[0]
    assert t.subject == v2.subject.incoming[0]
    assert t.subject.source == v1.subject
    assert t.subject.target == v2.subject


def test_final_state_head_glue(create):
    """Test transition head to final state connection."""
    v = create(FinalStateItem, UML.FinalState)
    t = create(TransitionItem)

    glued = allow(t, t.head, v)
    assert not glued



#!/usr/bin/env python3

from pathlib import Path
from tempfile import TemporaryDirectory

from dulwich.repo import Repo

from gaphor.storage.tests.fixtures import create_merge_conflict
from gaphor.ui import run

workspace = Path(__file__).parent.parent


def run_gaphor_with_merge_conflict():
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        repo = Repo.init(tmp_path)

        create_merge_conflict(
            repo,
            tmp_path / "RAAML.gaphor",
            (workspace / "test-models" / "RAAML-original.gaphor").read_text(),
            (workspace / "models" / "RAAML.gaphor").read_text(),
            (workspace / "test-models" / "RAAML-incoming.gaphor").read_text(),
        )

        run([__file__, str(tmp_path / "RAAML.gaphor")])


if __name__ == "__main__":
    run_gaphor_with_merge_conflict()



from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.profiles.stereotypepropertypages import (
    ShowStereotypePage,
    StereotypePage,
)


def create_property_page(diagram, element_factory, event_manager) -> StereotypePage:
    return StereotypePage(element_factory.create(UML.Class), event_manager)


def create_show_property_page(
    diagram, element_factory, event_manager
) -> ShowStereotypePage:
    item = diagram.create(
        UML.classes.ClassItem, subject=element_factory.create(UML.Class)
    )
    return ShowStereotypePage(item, event_manager)


def get_model(property_page: StereotypePage):
    widget = property_page.construct()
    show_stereotypes = find(widget, "stereotype-list")
    return show_stereotypes.get_model()


def metaclass_and_stereotype(element_factory):
    metaclass = element_factory.create(UML.Class)
    metaclass.name = "Class"
    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "Stereotype"
    UML.recipes.create_extension(metaclass, stereotype)
    return metaclass, stereotype


def test_show_stereotype_property_page_with_stereotype(
    diagram, element_factory, event_manager
):
    _metaclass, _stereotype = metaclass_and_stereotype(element_factory)
    property_page = create_show_property_page(diagram, element_factory, event_manager)

    widget = property_page.construct()
    show_stereotypes = find(widget, "show-stereotypes")
    show_stereotypes.set_active(True)

    assert property_page.item.show_stereotypes


def test_stereotype_property_page_apply_stereotype(
    diagram, element_factory, event_manager
):
    _metaclass, stereotype = metaclass_and_stereotype(element_factory)

    property_page = create_property_page(diagram, element_factory, event_manager)
    subject = property_page.subject
    model = get_model(property_page)
    view = model[0]
    view.applied = True

    assert stereotype in subject.appliedStereotype[0].classifier


def test_stereotype_property_page_slot_value(diagram, element_factory, event_manager):
    metaclass = element_factory.create(UML.Class)
    metaclass.name = "Class"
    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "Stereotype"
    stereotype.ownedAttribute = element_factory.create(UML.Property)
    UML.recipes.create_extension(metaclass, stereotype)

    property_page = create_property_page(diagram, element_factory, event_manager)
    model = get_model(property_page)
    model[0].applied = True
    model[1].slot_value = "test"
    applied_stereotype = property_page.subject.appliedStereotype[0]
    slot = applied_stereotype.slot[0]

    assert stereotype.ownedAttribute[0] is slot.definingFeature
    assert "test" == UML.recipes.get_slot_value(slot)


def test_inherited_stereotype(diagram, element_factory, event_manager):
    _metaclass, stereotype = metaclass_and_stereotype(element_factory)
    substereotype = element_factory.create(UML.Stereotype)
    substereotype.name = "SubStereotype"
    UML.recipes.create_generalization(stereotype, substereotype)

    property_page = create_property_page(diagram, element_factory, event_manager)
    model = get_model(property_page)

    assert model[0]
    assert model[0].name == "Stereotype"
    assert model[1]
    assert model[1].name == "SubStereotype"


def test_inherited_stereotype_with_attributes(diagram, element_factory, event_manager):
    _metaclass, stereotype = metaclass_and_stereotype(element_factory)
    attr = element_factory.create(UML.Property)
    attr.name = "Attr"
    stereotype.ownedAttribute = attr
    substereotype = element_factory.create(UML.Stereotype)
    substereotype.name = "SubStereotype"
    UML.recipes.create_generalization(stereotype, substereotype)

    property_page = create_property_page(diagram, element_factory, event_manager)
    model = get_model(property_page)

    assert model[0]
    assert model[0].name == "Stereotype"
    assert model[1].name == "Attr"
    assert model[2]
    assert model[2].name == "SubStereotype"
    assert model[3].name == "Attr"



import importlib

import pytest

from gaphor.main import main


def test_help_output(capsys):
    with pytest.raises(SystemExit, match="0"):
        main(["gaphor", "export", "--help"])

    captured = capsys.readouterr()
    assert "--help" in captured.out
    assert "--verbose" in captured.out
    assert "--use-underscores" in captured.out
    assert "--dir directory" in captured.out
    assert "--format format" in captured.out
    assert "--regex regex" in captured.out


@pytest.fixture
def model():
    return importlib.resources.files("test-models") / "all-elements.gaphor"


def test_export_pdf(tmp_path, model):
    main(["gaphor", "export", "-v", "-o", str(tmp_path), str(model)])

    model_path = tmp_path / "New model"

    assert model_path.exists()
    assert (model_path / "main.pdf").exists()


def test_export_png(tmp_path, model):
    main(["gaphor", "export", "-v", "-f", "png", "-o", str(tmp_path), str(model)])

    model_path = tmp_path / "New model"

    assert model_path.exists()
    assert (model_path / "main.png").exists()


def test_export_svg(tmp_path, model):
    main(["gaphor", "export", "-v", "-f", "svg", "-o", str(tmp_path), str(model)])

    model_path = tmp_path / "New model"

    assert model_path.exists()
    assert (model_path / "main.svg").exists()



import pytest

from gaphor.core.changeset.apply import applicable, apply_change
from gaphor.core.changeset.compare import set_value_change_property_value
from gaphor.core.modeling import (
    Diagram,
    ElementChange,
    RefChange,
    ValueChange,
)
from gaphor.UML import (
    Class,
    Element,
    LiteralBoolean,
    LiteralInteger,
    Property,
)
from gaphor.UML import Diagram as UMLDiagram


def test_create_element(element_factory, modeling_language):
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "12234"
    change.element_name = "Diagram"

    apply_change(change, element_factory, modeling_language)

    diagram = next(element_factory.select(Diagram))

    assert diagram
    assert element_factory.lookup(change.element_id)
    assert change.applied


def test_create_presentation(element_factory, modeling_language):
    diagram = element_factory.create(Diagram)
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "1234"
    change.element_name = "Box"
    change.diagram_id = diagram.id

    apply_change(change, element_factory, modeling_language)

    assert element_factory.lookup(change.element_id)
    assert change.applied


def test_create_presentation_without_diagram(element_factory, modeling_language):
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "12234"
    change.element_name = "Box"

    with pytest.raises(TypeError):
        apply_change(change, element_factory, modeling_language)


def test_remove_element(element_factory, modeling_language):
    diagram = element_factory.create(Diagram)
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "remove"
    change.element_id = diagram.id
    change.element_name = "Diagram"

    apply_change(change, element_factory, modeling_language)

    assert diagram not in element_factory
    assert change.applied


def test_update_str_value(element_factory, modeling_language):
    diagram = element_factory.create(Diagram)
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = diagram.id
    change.property_name = "name"
    set_value_change_property_value(change, "new value")

    apply_change(change, element_factory, modeling_language)

    assert diagram.name == "new value"
    assert change.applied


def test_update_bool_value(element_factory, modeling_language):
    literal = element_factory.create(LiteralBoolean)
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = literal.id
    change.property_name = "value"
    set_value_change_property_value(change, True)

    apply_change(change, element_factory, modeling_language)

    assert literal.value
    assert change.applied


def test_update_int_value(element_factory, modeling_language):
    literal = element_factory.create(LiteralInteger)
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = literal.id
    change.property_name = "value"
    set_value_change_property_value(change, 3)

    apply_change(change, element_factory, modeling_language)

    assert literal.value == 3
    assert change.applied


def test_update_str_default_value(element_factory, modeling_language):
    diagram = element_factory.create(Diagram)
    diagram.name = "old name"
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = diagram.id
    change.property_name = "name"

    apply_change(change, element_factory, modeling_language)

    assert not diagram.name
    assert change.applied


def test_update_bool_default_value(element_factory, modeling_language):
    klass = element_factory.create(Class)
    klass.isAbstract = True
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = klass.id
    change.property_name = "isAbstract"

    apply_change(change, element_factory, modeling_language)

    assert not klass.isAbstract
    assert change.applied


def test_update_enumeration_default_value(element_factory, modeling_language):
    property = element_factory.create(Property)
    property.aggregation = "shared"
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = property.id
    change.property_name = "aggregation"

    apply_change(change, element_factory, modeling_language)

    assert property.aggregation == "none"
    assert change.applied


def test_add_relation(element_factory, modeling_language):
    element = element_factory.create(Element)
    diagram = element_factory.create(UMLDiagram)
    change: RefChange = element_factory.create(RefChange)
    change.op = "add"
    change.element_id = diagram.id
    change.property_name = "element"
    change.property_ref = element.id

    apply_change(change, element_factory, modeling_language)

    assert diagram.element is element
    assert change.applied


def test_remove_single_relation(element_factory, modeling_language):
    element = element_factory.create(Element)
    diagram = element_factory.create(UMLDiagram)
    diagram.element = element

    change: RefChange = element_factory.create(RefChange)
    change.op = "remove"
    change.element_id = diagram.id
    change.property_name = "element"
    change.property_ref = element.id

    apply_change(change, element_factory, modeling_language)

    assert diagram.element is None


def test_remove_many_relation(element_factory, modeling_language):
    element = element_factory.create(Element)
    diagram = element_factory.create(UMLDiagram)
    diagram.element = element

    change: RefChange = element_factory.create(RefChange)
    change.op = "remove"
    change.element_id = element.id
    change.property_name = "ownedDiagram"
    change.property_ref = diagram.id

    apply_change(change, element_factory, modeling_language)

    assert diagram.element is None


def test_apply_opposite_relation(element_factory, modeling_language):
    element = element_factory.create(Element)
    diagram = element_factory.create(UMLDiagram)

    change: RefChange = element_factory.create(RefChange)
    change.op = "add"
    change.element_id = element.id
    change.property_name = "ownedDiagram"
    change.property_ref = diagram.id

    other: RefChange = element_factory.create(RefChange)
    other.op = "update"
    other.element_id = diagram.id
    other.property_name = "element"
    other.property_ref = element.id

    apply_change(change, element_factory, modeling_language)

    assert change.applied
    assert other.applied


def test_element_change_applicable(element_factory):
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "1234"
    change.element_name = "Presentation"

    assert applicable(change, element_factory)


def test_element_change_not_applicable(element_factory):
    element_factory.create_as(Element, id="1234")
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "1234"
    change.element_name = "Presentation"

    assert not applicable(change, element_factory)


def test_element_change_with_diagram_applicable(element_factory):
    diagram = element_factory.create(Diagram)
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "1234"
    change.element_name = "Presentation"
    change.diagram_id = diagram.id

    assert applicable(change, element_factory)


def test_element_change_without_diagram_not_applicable(element_factory):
    change: ElementChange = element_factory.create(ElementChange)
    change.op = "add"
    change.element_id = "1234"
    change.element_name = "Presentation"
    change.diagram_id = "does-not-exist"

    assert not applicable(change, element_factory)


def test_update_value_applicable(element_factory):
    diagram = element_factory.create(Diagram)
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = diagram.id
    change.property_name = "name"
    set_value_change_property_value(change, "new value")

    assert applicable(change, element_factory)


def test_update_value_not_applicable(element_factory):
    diagram = element_factory.create(Diagram)
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = diagram.id
    change.property_name = "name"
    set_value_change_property_value(change, "new value")
    diagram.name = "new value"

    assert not applicable(change, element_factory)


def test_update_value_without_element_not_applicable(element_factory):
    change: ValueChange = element_factory.create(ValueChange)
    change.element_id = "1234"
    change.property_name = "name"
    change.property_value = "new value"

    assert not applicable(change, element_factory)


def test_add_relation_applicable(element_factory, modeling_language):
    element = element_factory.create(Element)
    diagram = element_factory.create(Diagram)
    change: RefChange = element_factory.create(RefChange)
    change.op = "add"
    change.element_id = diagram.id
    change.property_name = "element"
    change.property_ref = element.id

    assert applicable(change, element_factory)



# ruff: noqa: SLF001
import pytest

from gaphor import UML
from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling import Base, ElementFactory
from gaphor.core.modeling.elementdispatcher import ElementDispatcher, EventWatcher
from gaphor.core.modeling.properties import association
from gaphor.UML.modelinglanguage import UMLModelingLanguage


class Event:
    def __init__(self):
        self.events = []

    def handler(self, event):
        self.events.append(event)


class A(Base):
    one: association
    two: association

    def __init__(self, id=None, model=None):
        super().__init__(id, model)


A.one = association("one", A, lower=0, upper=1, composite=True)
A.two = association("two", A, lower=0, upper=2, composite=True)


@pytest.fixture
def event_manager():
    return EventManager()


@pytest.fixture
def modeling_language():
    return UMLModelingLanguage()


@pytest.fixture
def dispatcher(event_manager, modeling_language):
    return ElementDispatcher(event_manager, modeling_language)


@pytest.fixture
def element_factory(event_manager, dispatcher):
    return ElementFactory(event_manager, dispatcher)


@pytest.fixture
def uml_class(element_factory):
    return element_factory.create(UML.Class)


@pytest.fixture
def uml_parameter(element_factory):
    return element_factory.create(UML.Parameter)


@pytest.fixture
def uml_operation(element_factory):
    return element_factory.create(UML.Operation)


@pytest.fixture
def uml_transition(element_factory):
    return element_factory.create(UML.Transition)


@pytest.fixture
def uml_constraint(element_factory):
    return element_factory.create(UML.Constraint)


@pytest.fixture
def event():
    return Event()


@pytest.fixture
def handler():
    events = []

    def handler(event):
        events.append(event)

    handler.events = events
    return handler


def test_register_handler(dispatcher, uml_class, uml_parameter, uml_operation, event):
    element = uml_class
    dispatcher.subscribe(event.handler, element, "ownedOperation.ownedParameter.name")
    assert len(dispatcher._handlers) == 1
    assert list(dispatcher._handlers.keys())[0] == (element, UML.Class.ownedOperation)

    # Add some properties:

    # 1:
    element.ownedOperation = uml_operation
    # 2:
    p = element.ownedOperation[0].ownedParameter = uml_parameter
    # 3:
    p.name = "func"
    dispatcher.subscribe(event.handler, element, "ownedOperation.ownedParameter.name")
    assert len(event.events) == 3
    assert len(dispatcher._handlers) == 3


def test_register_handler_twice(
    dispatcher, uml_class, uml_operation, uml_parameter, event
):
    """Multiple registrations have no effect."""
    # Add some properties:
    element = uml_class
    element.ownedOperation = uml_operation
    p = element.ownedOperation[0].ownedParameter = uml_parameter
    dispatcher.subscribe(event.handler, element, "ownedOperation.ownedParameter.name")

    n_handlers = len(dispatcher._handlers)

    assert len(event.events) == 0
    dispatcher.subscribe(event.handler, element, "ownedOperation.ownedParameter.name")
    assert n_handlers == len(dispatcher._handlers)
    dispatcher.subscribe(event.handler, element, "ownedOperation.ownedParameter.name")
    assert n_handlers == len(dispatcher._handlers)
    dispatcher.subscribe(event.handler, element, "ownedOperation.ownedParameter.name")
    assert n_handlers == len(dispatcher._handlers)

    p.name = "func"
    assert len(event.events) == 1


def test_unregister_handler(dispatcher, uml_class, uml_operation, uml_parameter, event):
    # First some setup:
    element = uml_class
    o = element.ownedOperation = uml_operation
    p = element.ownedOperation[0].ownedParameter = uml_parameter
    p.name = "func"
    dispatcher.subscribe(event.handler, element, "ownedOperation.ownedParameter.name")
    assert len(dispatcher._handlers) == 3
    assert dispatcher._handlers[element, UML.Class.ownedOperation]
    assert dispatcher._handlers[o, UML.Operation.ownedParameter]
    assert dispatcher._handlers[p, UML.Parameter.name]

    dispatcher.unsubscribe(event.handler)

    assert len(dispatcher._handlers) == 0, dispatcher._handlers
    assert len(dispatcher._reverse) == 0, dispatcher._reverse
    # Should not fail here too:
    dispatcher.unsubscribe(event.handler)


def test_notification(
    dispatcher, uml_class, uml_operation, uml_parameter, event, element_factory
):
    """Test notifications with Class object."""
    element = uml_class
    o = element.ownedOperation = uml_operation
    p = element.ownedOperation[0].ownedParameter = uml_parameter
    p.name = "func"
    dispatcher.subscribe(event.handler, element, "ownedOperation.ownedParameter.name")
    assert len(dispatcher._handlers) == 3
    assert not event.events

    element.ownedOperation = element_factory.create(UML.Operation)
    assert len(event.events) == 1, event.events
    assert len(dispatcher._handlers) == 4

    p.name = "othername"
    assert len(event.events) == 2, event.events

    del element.ownedOperation[o]
    assert len(dispatcher._handlers) == 2


def test_notification_2(
    dispatcher, uml_transition, uml_constraint, event, element_factory
):
    """Test notifications with Transition object."""
    element = uml_transition
    g = element.guard = uml_constraint
    dispatcher.subscribe(event.handler, element, "guard.specification")
    assert len(dispatcher._handlers) == 2
    assert not event.events
    assert (element.guard, UML.Constraint.specification) in list(
        dispatcher._handlers.keys()
    ), list(dispatcher._handlers.keys())

    specification = element_factory.create(UML.LiteralString)
    specification.owningConstraint = g
    specification.value = "x"
    g.specification = specification
    assert len(event.events) == 1, event.events

    element.guard = element_factory.create(UML.Constraint)
    assert len(event.events) == 2, event.events
    assert len(dispatcher._handlers) == 2, len(dispatcher._handlers)
    assert (element.guard, UML.Constraint.specification) in list(
        dispatcher._handlers.keys()
    )


def test_notification_of_change(
    dispatcher, uml_transition, uml_constraint, event, element_factory
):
    """Test notifications with Transition object."""
    element = uml_transition
    g = element.guard = uml_constraint
    dispatcher.subscribe(event.handler, element, "guard.specification")
    assert len(dispatcher._handlers) == 2
    assert not event.events

    specification = element_factory.create(UML.LiteralString)
    specification.owningConstraint = g
    specification.value = "x"
    g.specification = specification
    assert len(event.events) == 1, event.events

    element.guard = element_factory.create(UML.Constraint)
    assert len(event.events) == 2, event.events


def test_notification_with_composition(
    dispatcher, uml_class, uml_operation, uml_constraint, event
):
    """Test unregister with composition.

    Use Class.ownedOperation.precondition.
    """
    element = uml_class
    o = element.ownedOperation = uml_operation
    p = element.ownedOperation[0].precondition = uml_constraint
    p.name = "func"
    dispatcher.subscribe(event.handler, element, "ownedOperation.precondition.name")
    assert len(dispatcher._handlers) == 3
    assert not event.events

    del element.ownedOperation[o]
    assert len(dispatcher._handlers) == 1


def test_notification_with_incompatible_elements(
    dispatcher, uml_transition, uml_constraint, event
):
    """Test unregister with composition.

    Use Class.ownedOperation.precondition.
    """
    element = uml_transition
    g = element.guard = uml_constraint
    dispatcher.subscribe(event.handler, element, "guard.specification")
    assert len(dispatcher._handlers) == 2
    assert not event.events
    assert (element.guard, UML.Constraint.specification) in list(
        dispatcher._handlers.keys()
    ), list(dispatcher._handlers.keys())

    specification = uml_transition.model.create(UML.LiteralString)
    specification.owningConstraint = g
    specification.value = "x"
    g.specification = specification
    assert len(event.events) == 1, event.events

    specification = uml_transition.model.create(UML.LiteralString)
    specification.owningConstraint = g
    specification.value = "a"
    g.specification = specification
    g.specification.value = "a"
    assert len(event.events) == 2, event.events


def test_notification_with_class(element_factory, dispatcher, handler):
    """Test notifications with Class object."""
    element = element_factory.create(UML.Class)
    o = element.ownedOperation = element_factory.create(UML.Operation)
    p = element.ownedOperation[0].ownedParameter = element_factory.create(UML.Parameter)
    p.name = "func"
    dispatcher.subscribe(handler, element, "ownedOperation.ownedParameter.name")
    assert len(dispatcher._handlers) == 3
    assert not handler.events

    element.ownedOperation = element_factory.create(UML.Operation)
    assert len(handler.events) == 1, handler.events
    assert len(dispatcher._handlers) == 4

    p.name = "othername"
    assert len(handler.events) == 2, handler.events

    del element.ownedOperation[o]
    assert len(dispatcher._handlers) == 2


def test_association_notification(element_factory, dispatcher, handler):
    """Test notifications with Class object.

    Tricky case where no events are fired.
    """
    element = element_factory.create(UML.Association)
    p1 = element.memberEnd = element_factory.create(UML.Property)
    element.memberEnd = element_factory.create(UML.Property)

    assert len(element.memberEnd) == 2
    dispatcher.subscribe(handler, element, "memberEnd.name")
    assert len(dispatcher._handlers) == 3, len(dispatcher._handlers)
    assert not handler.events

    p1.name = "foo"
    assert len(handler.events) == 1, (handler.events, dispatcher._handlers)
    assert len(dispatcher._handlers) == 3

    p1.name = "othername"
    assert len(handler.events) == 2, handler.events

    p1.name = "othername"
    assert len(handler.events) == 2, handler.events


def test_association_notification_complex(element_factory, dispatcher, handler):
    """Test notifications with Class object.

    Tricky case where no events are fired.
    """
    element = element_factory.create(UML.Association)
    p1 = element.memberEnd = element_factory.create(UML.Property)
    p2 = element.memberEnd = element_factory.create(UML.Property)
    UML.recipes.set_multiplicity_lower_value(p1, "0")
    UML.recipes.set_multiplicity_upper_value(p1, "1")
    UML.recipes.set_multiplicity_lower_value(p2, "1")
    UML.recipes.set_multiplicity_upper_value(p2, "*")

    assert len(element.memberEnd) == 2

    base = "memberEnd[Property]."
    dispatcher.subscribe(handler, element, f"{base}name")
    dispatcher.subscribe(handler, element, f"{base}aggregation")
    dispatcher.subscribe(handler, element, f"{base}classifier")
    dispatcher.subscribe(handler, element, f"{base}lowerValue")
    dispatcher.subscribe(handler, element, f"{base}upperValue")

    assert len(dispatcher._handlers) == 11, len(dispatcher._handlers)
    assert not handler.events

    p1.name = "foo"
    assert len(handler.events) == 1, (handler.events, dispatcher._handlers)
    assert len(dispatcher._handlers) == 11

    p1.name = "othername"
    assert len(handler.events) == 2, handler.events


def test_diamond(element_factory, dispatcher, handler):
    """Test diamond shaped dependencies a -> b -> c, a -> b' -> c."""
    a = element_factory.create(A)
    watcher = EventWatcher(a, dispatcher, handler)
    watcher.watch("one.two.one.two")

    a.one = element_factory.create(A)
    a.one.two = element_factory.create(A)
    a.one.two = element_factory.create(A)
    a.one.two[0].one = element_factory.create(A)
    a.one.two[1].one = a.one.two[0].one
    a.one.two[0].one.two = element_factory.create(A)

    assert len(handler.events) == 6

    a.unlink()
    watcher.unsubscribe_all()
    watcher.unsubscribe_all()


def test_big_diamond(element_factory, dispatcher, handler):
    """Test diamond shaped dependencies a -> b -> c -> d, a -> b' -> c' ->
    d."""
    a = element_factory.create(A)
    watcher = EventWatcher(a, dispatcher, handler)
    watcher.watch("one.two.one.two")

    a.one = element_factory.create(A)
    a.one.two = element_factory.create(A)
    a.one.two = element_factory.create(A)
    a.one.two[0].one = element_factory.create(A)
    a.one.two[1].one = element_factory.create(A)
    a.one.two[0].one.two = element_factory.create(A)
    a.one.two[1].one.two = a.one.two[0].one.two[0]

    assert len(handler.events) == 7

    a.unlink()
    watcher.unsubscribe_all()
    watcher.unsubscribe_all()
    assert len(dispatcher._handlers) == 0


def test_braking_big_diamond(element_factory, dispatcher, handler):
    """Test diamond shaped dependencies a -> b -> c -> d, a -> b' -> c' ->
    d."""
    a = element_factory.create(A)
    watcher = EventWatcher(a, dispatcher, handler)
    watcher.watch("one.two.one.two")

    a.one = element_factory.create(A)
    a.one.two = element_factory.create(A)
    a.one.two = element_factory.create(A)
    a.one.two[0].one = element_factory.create(A)
    a.one.two[1].one = element_factory.create(A)
    a.one.two[0].one.two = element_factory.create(A)
    a.one.two[1].one.two = a.one.two[0].one.two[0]

    assert len(handler.events) == 7
    assert len(dispatcher._handlers) == 6

    del a.one.two[0].one
    watcher.unsubscribe_all()
    watcher.unsubscribe_all()
    assert len(dispatcher._handlers) == 0


def test_cyclic(element_factory, dispatcher, handler):
    """Test cyclic dependency a -> b -> c -> a."""
    a = element_factory.create(A)
    watcher = EventWatcher(a, dispatcher, handler)
    watcher.watch("one.two.one.two")

    a.one = element_factory.create(A)
    a.one.two = element_factory.create(A)
    a.one.two = element_factory.create(A)
    a.one.two[0].one = a

    assert 4 == len(handler.events)

    a.unlink()
    assert 1 == len(dispatcher._handlers)



import pytest

from gaphor.core.modeling.properties import attribute
from gaphor.SysML.diagramtype import DiagramDefault, SysMLDiagramType
from gaphor.SysML.sysml import SysMLDiagram
from gaphor.UML import NamedElement


class MockElementA(NamedElement):
    pass


class MockElementB(NamedElement):
    pass


class AbcSysMLDiagram(SysMLDiagram):
    diagramType: attribute[str] = attribute("diagramType", str, default="abc")


@pytest.fixture
def diagram_type():
    return SysMLDiagramType(
        AbcSysMLDiagram,
        "Defghi",
        (),
        (MockElementA,),
        (DiagramDefault(type(None), MockElementA, "New mock element A"),),
    )


def test_sysml_diagram_type_allowed(diagram_type):
    assert diagram_type.allowed(MockElementA())
    assert diagram_type.allowed(None)
    assert not diagram_type.allowed(MockElementB())


def test_sysml_diagram_type_with_parent(element_factory, diagram_type):
    mock_a = element_factory.create(MockElementA)
    mock_a.name = "Mock A"

    diagram = diagram_type.create(element_factory, mock_a)
    assert diagram.diagramType == "abc"
    assert diagram.name == "Defghi"
    assert isinstance(diagram.element, MockElementA)
    assert diagram.element.name == "Mock A"


def test_sysml_diagram_type_without_parent(element_factory, diagram_type):
    diagram = diagram_type.create(element_factory, None)
    assert diagram.diagramType == "abc"
    assert diagram.name == "Defghi"
    assert isinstance(diagram.element, MockElementA)
    assert diagram.element.name == "New mock element A"


def test_sysml_diagram_type_wrong_parent(element_factory, diagram_type):
    mock_b = element_factory.create(MockElementB)
    mock_b.name = "Mock B"

    with pytest.raises(TypeError) as e:
        diagram_type.create(element_factory, mock_b)
    assert str(e.value) == "Can’t create “Defghi” nested under a “Mock B”"



from gaphor.diagram.drop import drop
from gaphor.UML import Action, ControlFlow, ObjectFlow, ObjectNode, uml
from gaphor.UML.actions.action import ActionItem
from gaphor.UML.actions.actionstoolbox import partition_config
from gaphor.UML.actions.activity import ActivityItem, ActivityParameterNodeItem
from gaphor.UML.actions.drop import drop_on_partition
from gaphor.UML.actions.partition import PartitionItem


def test_activity_parameter_drop_with_single_activity_and_parameter(
    diagram, element_factory
):
    activity = element_factory.create(uml.Activity)
    activity_item = diagram.create(ActivityItem, subject=activity)

    activity_parameter_node = element_factory.create(uml.ActivityParameterNode)
    activity.node = activity_parameter_node

    parameter_item = drop(activity_parameter_node, diagram, 0, 0)
    assert isinstance(parameter_item, ActivityParameterNodeItem)
    assert parameter_item.subject is activity_parameter_node
    assert parameter_item.parent is activity_item


def test_activity_parameter_drop_with_multiple_activities_and_parameters(
    diagram, element_factory
):
    activity_1 = element_factory.create(uml.Activity)
    activity_2 = element_factory.create(uml.Activity)
    activity_item_1 = diagram.create(ActivityItem, subject=activity_1)
    activity_item_2 = diagram.create(ActivityItem, subject=activity_2)

    activity_parameter_node_1 = element_factory.create(uml.ActivityParameterNode)
    activity_parameter_node_2 = element_factory.create(uml.ActivityParameterNode)
    activity_1.node = activity_parameter_node_1
    activity_2.node = activity_parameter_node_2

    parameter_item_1 = drop(activity_parameter_node_1, diagram, 0, 0)
    parameter_item_2 = drop(activity_parameter_node_2, diagram, 20, 30)

    assert isinstance(parameter_item_1, ActivityParameterNodeItem)
    assert isinstance(parameter_item_2, ActivityParameterNodeItem)
    assert parameter_item_1.subject is activity_parameter_node_1
    assert parameter_item_2.subject is activity_parameter_node_2
    assert parameter_item_1.parent is activity_item_1
    assert parameter_item_2.parent is activity_item_2


def test_activity_parameter_drop_with_two_same_activity_items(diagram, element_factory):
    activity = element_factory.create(uml.Activity)
    activity_item_1 = diagram.create(ActivityItem, subject=activity)
    activity_item_2 = diagram.create(ActivityItem, subject=activity)
    activity_item_1.matrix.translate(50, 50)
    activity_item_2.matrix.translate(300, -200)

    activity_parameter_node = element_factory.create(uml.ActivityParameterNode)
    activity.node = activity_parameter_node

    parameter_item_1 = drop(activity_parameter_node, diagram, 0, 0)
    parameter_item_2 = drop(activity_parameter_node, diagram, 320, -230)

    assert isinstance(parameter_item_1, ActivityParameterNodeItem)
    assert parameter_item_1.subject is activity_parameter_node
    assert parameter_item_1.parent is activity_item_1

    assert isinstance(parameter_item_2, ActivityParameterNodeItem)
    assert parameter_item_2.subject is activity_parameter_node
    assert parameter_item_2.parent is activity_item_2


def test_drop_control_flow(diagram, element_factory):
    a = element_factory.create(Action)
    b = element_factory.create(Action)
    control_flow = element_factory.create(ControlFlow)
    control_flow.source = a
    control_flow.target = b

    drop(a, diagram, 0, 0)
    drop(b, diagram, 0, 0)
    item = drop(control_flow, diagram, 0, 0)

    assert item
    assert diagram.connections.get_connection(item.head).connected is a.presentation[0]
    assert diagram.connections.get_connection(item.tail).connected is b.presentation[0]


def test_drop_object_flow(diagram, element_factory):
    a = element_factory.create(Action)
    b = element_factory.create(ObjectNode)
    control_flow = element_factory.create(ObjectFlow)
    control_flow.source = a
    control_flow.target = b

    drop(a, diagram, 0, 0)
    drop(b, diagram, 0, 0)
    item = drop(control_flow, diagram, 0, 0)

    assert item
    assert diagram.connections.get_connection(item.head).connected is a.presentation[0]
    assert diagram.connections.get_connection(item.tail).connected is b.presentation[0]


def test_drop_on_swimlane_on_first_partition(create):
    swimlanes: PartitionItem = create(PartitionItem, uml.ActivityPartition)
    partition_config(swimlanes)
    action = create(ActionItem, uml.Action)

    drop_on_partition(action, swimlanes, 10, swimlanes.height / 2)

    assert action.subject in swimlanes.partition[0].node


def test_drop_on_swimlane_on_second_partition(create):
    swimlanes: PartitionItem = create(PartitionItem, uml.ActivityPartition)
    partition_config(swimlanes)
    action = create(ActionItem, uml.Action)

    drop_on_partition(action, swimlanes, swimlanes.width - 10, swimlanes.height / 2)

    assert action.subject in swimlanes.partition[1].node


def test_drop_from_first_swimlane_on_second_partition(create):
    swimlanes: PartitionItem = create(PartitionItem, uml.ActivityPartition)
    partition_config(swimlanes)
    action = create(ActionItem, uml.Action)

    handles = swimlanes.handles()
    drop_on_partition(action, swimlanes, handles[0].pos.x + 10, swimlanes.height / 2)
    drop_on_partition(action, swimlanes, handles[1].pos.x - 10, swimlanes.height / 2)

    assert action.subject not in swimlanes.partition[0].node
    assert action.subject in swimlanes.partition[1].node


def test_drop_from_second_swimlane_to_outside(create):
    swimlanes: PartitionItem = create(PartitionItem, uml.ActivityPartition)
    partition_config(swimlanes)
    action = create(ActionItem, uml.Action)

    handles = swimlanes.handles()
    drop_on_partition(action, swimlanes, handles[1].pos.x - 10, swimlanes.height / 2)
    drop_on_partition(action, swimlanes, handles[1].pos.x - 10, 1000)

    assert action.subject not in swimlanes.partition[0].node
    assert action.subject not in swimlanes.partition[1].node


def test_move_action_from_activity_to_swimlane(create):
    activity: ActivityItem = create(ActivityItem, uml.Activity)
    swimlanes: PartitionItem = create(PartitionItem, uml.ActivityPartition)
    partition_config(swimlanes)
    action = create(ActionItem, uml.Action)

    drop(action, activity, 0, 0)
    drop(swimlanes, activity, 0, 0)

    handles = swimlanes.handles()
    drop_on_partition(action, swimlanes, handles[0].pos.x + 10, swimlanes.height / 2)

    assert action.subject in swimlanes.partition[0].node
    assert action.subject not in activity.subject.ownedElement



from gi.repository import Gtk

from gaphor.diagram.tools.txtool import TxData, on_begin, on_end, transactional_tool
from gaphor.transaction import TransactionBegin


class MockEventManager:
    def __init__(self):
        self.events = []

    def handle(self, event):
        self.events.append(event)


def test_start_tx_on_begin(view):
    event_manager = MockEventManager()
    tx_data = TxData(event_manager)
    (tool, _key_tool) = transactional_tool(
        Gtk.GestureDrag.new(), event_manager=event_manager
    )  # type: ignore[arg-type]

    on_begin(tool, None, tx_data)
    assert tx_data.txs

    on_end(tool, None, tx_data)

    assert event_manager.events
    assert isinstance(event_manager.events[0], TransactionBegin)
    assert not tx_data.txs



# ruff: noqa: F401,F811
from gaphor import UML
from gaphor.conftest import (
    create,
    diagram,
    element_factory,
    event_manager,
    modeling_language,
)
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.classes import AssociationItem, ClassItem
from gaphor.UML.classes.associationpropertypages import AssociationPropertyPage
from gaphor.UML.classes.classestoolbox import composite_association_config


def test_connect_composite_association(create, event_manager):
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    a = create(AssociationItem)
    composite_association_config(a)

    property_page = AssociationPropertyPage(a.subject, event_manager)
    _widget = property_page.construct()

    connect(a, a.head, c1)
    connect(a, a.tail, c2)



from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.states.propertypages import (
    RegionPropertyPage,
    StatePropertyPage,
    TransitionPropertyPage,
)
from gaphor.UML.states.state import StateItem


def test_state_property_page_entry(element_factory, event_manager):
    # Create some behavior elements
    opt1 = element_factory.create(UML.Activity)
    opt1.name = "option 1"

    opt2 = element_factory.create(UML.Interaction)
    opt2.name = "option 2"

    # Create subject property page
    subject = element_factory.create(UML.State)
    property_page = StatePropertyPage(subject, event_manager)

    widget = property_page.construct()
    do_activity = find(widget, "entry")

    do_activity.set_selected(0)
    assert not subject.entry

    do_activity.set_selected(1)
    assert (
        subject.entry
        and isinstance(subject.entry, UML.Activity)
        and subject.entry.name == "option 1"
    )

    do_activity.set_selected(2)
    assert (
        subject.entry
        and isinstance(subject.entry, UML.Interaction)
        and subject.entry.name == "option 2"
    )

    do_activity.set_selected(0)
    assert not subject.entry


def test_state_property_page_exit(element_factory, event_manager):
    # Create some behavior elements
    opt1 = element_factory.create(UML.Activity)
    opt1.name = "option 1"

    opt2 = element_factory.create(UML.Interaction)
    opt2.name = "option 2"

    # Create subject property page
    subject = element_factory.create(UML.State)
    property_page = StatePropertyPage(subject, event_manager)

    widget = property_page.construct()
    do_activity = find(widget, "exit")

    do_activity.set_selected(0)
    assert not subject.exit

    do_activity.set_selected(1)
    assert (
        subject.exit
        and isinstance(subject.exit, UML.Activity)
        and subject.exit.name == "option 1"
    )

    do_activity.set_selected(2)
    assert (
        subject.exit
        and isinstance(subject.exit, UML.Interaction)
        and subject.exit.name == "option 2"
    )

    do_activity.set_selected(0)
    assert not subject.exit


def test_state_property_page_do_activity(element_factory, event_manager):
    # Create some behavior elements
    opt1 = element_factory.create(UML.Activity)
    opt1.name = "option 1"

    opt2 = element_factory.create(UML.Interaction)
    opt2.name = "option 2"

    # Create subject property page
    subject = element_factory.create(UML.State)
    property_page = StatePropertyPage(subject, event_manager)

    widget = property_page.construct()
    do_activity = find(widget, "do-activity")

    do_activity.set_selected(0)
    assert not subject.doActivity

    do_activity.set_selected(1)
    assert (
        subject.doActivity
        and isinstance(subject.doActivity, UML.Activity)
        and subject.doActivity.name == "option 1"
    )

    do_activity.set_selected(2)
    assert (
        subject.doActivity
        and isinstance(subject.doActivity, UML.Interaction)
        and subject.doActivity.name == "option 2"
    )

    do_activity.set_selected(0)
    assert not subject.doActivity


def test_transition_property_page(element_factory, event_manager):
    subject = element_factory.create(UML.Transition)
    property_page = TransitionPropertyPage(subject, event_manager)

    widget = property_page.construct()
    guard = find(widget, "guard")
    guard.set_text("test")

    assert subject.guard.specification.value == "test"


def test_region_property_page(create, event_manager):
    item = create(StateItem, UML.State)
    property_page = RegionPropertyPage(item, event_manager)

    widget = property_page.construct()
    num_regions = find(widget, "num-regions")
    num_regions.set_value(4)

    assert len(item.subject.region) == 4



# ruff: noqa T201
#
# Find the latest (top most) version from the appdata.

from xml.etree import ElementTree as etree

tree = etree.parse("data/org.gaphor.Gaphor.appdata.xml")

if (release := tree.find("./releases/release")) is not None:
    print(release.attrib["version"])



from gaphor import UML
from gaphor.diagram.tests.fixtures import (
    connect,
    copy_and_paste_link,
    copy_clear_and_paste_link,
)
from gaphor.UML.classes import ClassItem
from gaphor.UML.profiles.extension import ExtensionItem


def metaclass_stereotype_and_extension(diagram, element_factory):
    m_cls = element_factory.create(UML.Class)
    st_cls = element_factory.create(UML.Stereotype)
    m_cls_item = diagram.create(ClassItem, subject=m_cls)
    st_cls_item = diagram.create(ClassItem, subject=st_cls)
    ext_item = diagram.create(ExtensionItem)
    m_cls.name = "Class"
    st_cls.name = "Stereotype"

    connect(ext_item, ext_item.handles()[0], m_cls_item)
    connect(ext_item, ext_item.handles()[1], st_cls_item)

    return m_cls_item, st_cls_item, ext_item


def test_copy_paste_of_stereotype(diagram, element_factory):
    m_cls_item, st_cls_item, ext_item = metaclass_stereotype_and_extension(
        diagram, element_factory
    )

    copy_and_paste_link({m_cls_item, ext_item, st_cls_item}, diagram, element_factory)

    new_ext_item = ext_item.subject.presentation[1]

    assert new_ext_item.head
    assert new_ext_item.tail


def test_cut_paste_of_stereotype(diagram, element_factory):
    m_cls_item, st_cls_item, ext_item = metaclass_stereotype_and_extension(
        diagram, element_factory
    )

    copy_clear_and_paste_link(
        {m_cls_item, st_cls_item, ext_item}, diagram, element_factory
    )
    new_m_cls = next(element_factory.select(UML.Class))
    new_st_cls = next(element_factory.select(UML.Stereotype))
    new_ext = next(element_factory.select(UML.Extension))

    assert new_m_cls in new_ext.memberEnd[:].type
    assert new_st_cls in new_ext.memberEnd[:].type
    assert new_st_cls is new_ext.ownedEnd.type



from gaphor import UML
from gaphor.core.modeling.collection import collection
from gaphor.core.modeling.diagram import (
    Diagram,
    StyledItem,
    attrname,
    lookup_attribute,
)
from gaphor.UML.classes import ClassItem


def test_attribute_of_item(diagram):
    classitem = diagram.create(ClassItem)

    node = StyledItem(classitem)

    assert node.attribute("show_attributes") == "true"


def test_nonexistent_attribute_of_item(diagram):
    classitem = diagram.create(ClassItem)

    node = StyledItem(classitem)

    assert node.attribute("foobar") is None


def test_should_get_type_of_element_attribute(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    classitem = diagram.create(ClassItem, subject=class_)

    node = StyledItem(classitem)

    assert node.attribute("subject") == "class"


def test_should_read_attribute_of_subject(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    classitem = diagram.create(ClassItem, subject=class_)

    class_.name = "myname"
    node = StyledItem(classitem)

    assert node.attribute("name") == "myname"


# test attributes that can be any of a list
def test_nested_attribute_of_subject(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    attr: UML.Property = element_factory.create(UML.Property)
    class_.ownedAttribute = attr
    classitem = diagram.create(ClassItem, subject=class_)

    attr.name = "myname"
    attr.isStatic = True
    node = StyledItem(classitem)

    assert node.attribute("attribute") == "property"
    assert node.attribute("attribute.name") == "myname"
    assert node.attribute("attribute.ownedElement") == ""
    assert node.attribute("attribute.isStatic") == "true"
    assert node.attribute("attribute.notAnAttribute") is None


def test_multiple_nested_attribute_of_subject(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    attr1 = element_factory.create(UML.Property)
    attr2 = element_factory.create(UML.Property)
    class_.ownedAttribute = attr1
    class_.ownedAttribute = attr2
    classitem = diagram.create(ClassItem, subject=class_)

    attr1.name = "first"
    attr2.name = "second"
    node = StyledItem(classitem)

    assert node.attribute("attribute") == "property property"
    assert node.attribute("attribute.name") in ("first second", "second first")
    assert node.attribute("attribute.notAnAttribute") is None


def test_mixed_case_attributes(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    attr1 = element_factory.create(UML.Property)
    class_.ownedAttribute = attr1
    classitem = diagram.create(ClassItem, subject=class_)

    attr1.name = "first"
    node = StyledItem(classitem)

    assert node.attribute("ownedattribute") == "property"
    assert node.attribute("ownedattribute.name") == "first"


def test_get_existing_attribute():
    diagram = Diagram()
    diagram.diagramType = "Test"

    assert diagram.diagramType == "Test"
    assert lookup_attribute(diagram, "diagramType") == "test"


def test_non_existent_attribute():
    diagram = Diagram()

    assert lookup_attribute(diagram, "doesnotexist") is None


def test_nested_attribute():
    diagram = UML.Diagram()
    diagram.ownedDiagram = UML.Diagram()

    assert lookup_attribute(diagram, "ownedDiagram.name") == ""
    assert lookup_attribute(diagram, "ownedDiagram.doesnotexist") is None


def test_non_existent_nested_attribute():
    diagram = Diagram()

    assert lookup_attribute(diagram, "ownedDiagram.name") is None
    assert lookup_attribute(diagram, "ownedDiagram.doesnotexist") is None


def test_attrname_diagram_subject():
    diagram = Diagram()
    assert attrname(diagram, "subject") == "subject"


def test_attrname_collection_subject(diagram):
    collection1 = collection(None, None, int)
    assert attrname(collection1, "subject") == "subject"


def test_attribute_on_item_and_not_on_subject(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    classitem = diagram.create(ClassItem, subject=class_)

    node = StyledItem(classitem)

    assert node.attribute("children") == ""
    assert node.attribute("does_not_exist") is None



from gaphor.core.modeling import ElementFactory
from gaphor.diagram.group import can_group, change_owner, group, ungroup
from gaphor.UML import Diagram, Element


def test_group_diagram(element_factory):
    diagram = element_factory.create(Diagram)
    parent = element_factory.create(Element)

    assert group(parent, diagram)

    assert diagram.element is parent


def test_ungroup_diagram(element_factory):
    diagram = element_factory.create(Diagram)
    parent = element_factory.create(Element)
    diagram.element = parent

    assert ungroup(parent, diagram)

    assert diagram.element is None


def test_ungroup_no_parent(element_factory):
    element = element_factory.create(Element)

    assert ungroup(None, element)


def test_do_not_ungroup_diagram_from_wrong_parent(element_factory):
    diagram = element_factory.create(Diagram)
    parent = element_factory.create(Element)
    wrong_parent = element_factory.create(Element)
    diagram.element = parent

    assert not ungroup(wrong_parent, diagram)

    assert diagram.element is parent


def test_ungroup_with_none(element_factory):
    parent = element_factory.create(Element)

    assert ungroup(parent, None)


def test_can_group_with_type(element_factory):
    parent = element_factory.create(Element)

    assert can_group(parent, Diagram)


def test_can_group_with_instance(element_factory):
    diagram = element_factory.create(Diagram)
    parent = element_factory.create(Element)

    assert can_group(parent, diagram)


def test_cannot_change_owner_from_different_models(element_factory):
    other_element_factory = ElementFactory()
    diagram = element_factory.create(Diagram)
    parent = other_element_factory.create(Element)

    assert not change_owner(parent, diagram)


def test_change_owner_to_root(element_factory):
    diagram = element_factory.create(Diagram)
    parent = element_factory.create(Element)
    change_owner(parent, diagram)

    assert change_owner(None, diagram)
    assert diagram.owner is None



"""Test connection of containment relationship."""

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.UML.classes import ClassItem, PackageItem
from gaphor.UML.classes.containment import ContainmentItem


def test_containment_package_glue(create):
    """Test containment glue to two package items."""
    pkg1 = create(PackageItem, UML.Package)
    containment = create(ContainmentItem)

    assert allow(containment, containment.head, pkg1)


def test_containment_package_glue_connected_on_one_end(create):
    """Test containment glue to two package items."""
    pkg1 = create(PackageItem, UML.Package)
    pkg2 = create(PackageItem, UML.Package)
    containment = create(ContainmentItem)

    connect(containment, containment.head, pkg1)

    assert allow(containment, containment.tail, pkg2)
    assert not allow(containment, containment.tail, pkg1)


def test_containment_can_not_create_cycles(create, diagram, element_factory):
    """Test containment connecting to a package and a class."""
    package = create(PackageItem, UML.Package)
    klass = create(ClassItem, UML.Class)
    klass.subject.package = package.subject
    line = create(ContainmentItem)

    connect(line, line.head, klass)

    assert not allow(line, line.tail, package)


def test_containment_package_class(create, diagram):
    """Test containment connecting to a package and a class."""
    package = create(PackageItem, UML.Package)
    line = create(ContainmentItem)
    klass = create(ClassItem, UML.Class)

    connect(line, line.head, package)
    connect(line, line.tail, klass)

    assert diagram.connections.get_connection(line.tail).connected is klass
    assert len(package.subject.ownedElement) == 1
    assert klass.subject in package.subject.ownedElement


def test_containment_package_class_disconnect(create, diagram, element_factory):
    """Test containment disconnecting from a package and a class."""
    parent_package = element_factory.create(UML.Package)
    diagram.element = parent_package

    package = create(PackageItem, UML.Package)
    line = create(ContainmentItem)
    klass = create(ClassItem, UML.Class)

    connect(line, line.tail, klass)
    connect(line, line.head, package)
    disconnect(line, line.head)

    assert klass.subject in parent_package.ownedElement


def test_containment_class_class(create, diagram, element_factory):
    """Test containment connecting to a package and a class."""
    parent_package = element_factory.create(UML.Package)
    container = create(ClassItem, UML.Class)
    container.subject.package = parent_package
    line = create(ContainmentItem)
    klass = create(ClassItem, UML.Class)
    klass.subject.package = parent_package

    connect(line, line.head, container)
    connect(line, line.tail, klass)

    assert diagram.connections.get_connection(line.tail).connected is klass
    assert len(container.subject.ownedElement) == 1
    assert klass.subject.owner is container.subject
    assert klass.subject in container.subject.ownedElement


def test_containment_class_class_disconnect(create, diagram, element_factory):
    """Test containment connecting to a package and a class."""
    parent_package = element_factory.create(UML.Package)
    diagram.element = parent_package

    container = create(ClassItem, UML.Class)
    line = create(ContainmentItem)
    klass = create(ClassItem, UML.Class)

    connect(line, line.head, container)
    connect(line, line.tail, klass)
    disconnect(line, line.head)

    assert klass.subject.owner is parent_package
    assert klass.subject in parent_package.ownedElement


def test_containment_reconnect_in_new_diagram(create, element_factory):
    # Most recently created containment relation wins.
    rel = create(ContainmentItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(rel, rel.head, c1)
    connect(rel, rel.tail, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(UML.Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    rel2 = diagram2.create(ContainmentItem)

    connect(rel2, rel2.head, c3)
    connect(rel2, rel2.tail, c4)

    c5 = diagram2.create(ClassItem, subject=element_factory.create(UML.Class))
    connect(rel2, rel2.head, c5)

    assert c2.subject.owner is c5.subject
    assert c4.subject.owner is c5.subject



from gaphor.UML.actions.pin import InputPinItem


def test_load_pin(element_factory, diagram, saver, loader):
    attached = diagram.create(InputPinItem)

    attached.handles()[0].pos = (1, 2)
    attached.matrix.translate(10, 20)

    dump = saver()
    loader(dump)

    new_attached = next(element_factory.select(InputPinItem))

    assert tuple(map(float, new_attached.handles()[0].pos)) == (1, 2)
    assert tuple(new_attached.matrix) == (1.0, 0.0, 0.0, 1.0, 10, 20)



import logging
import os
import sys

from gaphor.ui import run


def test_self_test(caplog):
    caplog.set_level(logging.DEBUG)
    exit_code = run(argv=[], launch_service="self_test")

    # Self test tends to fail every now and then on macOS.
    # We have the same issue for the final package check,
    # but a retry normally fixes it.
    # So let's retry on GitHub Actions on macOS only.
    if exit_code != 0 and sys.platform == "darwin" and os.getenv("CI"):
        exit_code = run(argv=[], launch_service="self_test")

    assert exit_code == 0



import pytest

from gaphor import UML
from gaphor.application import Session
from gaphor.core import Transaction
from gaphor.core.modeling import Diagram
from gaphor.storage.load import load


@pytest.fixture
def session():
    session = Session()
    yield session
    session.shutdown()


@pytest.fixture
def event_manager(session):
    return session.get_service("event_manager")


@pytest.fixture
def element_factory(session, test_models):
    element_factory = session.get_service("element_factory")
    modeling_language = session.get_service("modeling_language")
    with (test_models / "issue_53.gaphor").open(encoding="utf-8") as f:
        load(f, element_factory, modeling_language)
    yield element_factory
    element_factory.shutdown()


def test_package_removal(event_manager, element_factory):
    # Find all profile instances
    profiles = element_factory.lselect(UML.Profile)

    # Check there is 1 profile
    assert len(profiles) == 1

    # Check the profile has 1 presentation
    assert len(profiles[0].presentation) == 1

    # Unlink the presentation and profile
    with Transaction(event_manager):
        profiles[0].presentation[0].unlink()
        profiles[0].unlink()

    assert not element_factory.lselect(UML.Profile)

    classes = element_factory.lselect(UML.Class)
    assert len(classes) == 1

    # Check if the link is really removed:
    assert not classes[0].appliedStereotype
    assert not element_factory.lselect(UML.InstanceSpecification)
    assert len(element_factory.lselect(Diagram)) == 3


def test_package_removal_by_removing_the_diagram(event_manager, element_factory):
    diagram = element_factory.lselect(
        lambda e: e.isKindOf(Diagram) and e.name == "Stereotypes diagram"
    )[0]

    assert diagram

    with Transaction(event_manager):
        diagram.unlink()

    assert not element_factory.lselect(UML.Profile)

    classes = element_factory.lselect(UML.Class)
    assert len(classes) == 1

    # Check if the link is really removed:
    assert not classes[0].appliedStereotype
    assert not element_factory.lselect(UML.InstanceSpecification)
    assert len(element_factory.lselect(Diagram)) == 2



"""Message connection adapter tests."""

import pytest

from gaphor import UML
from gaphor.diagram.group import group
from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.UML.interactions.executionspecification import ExecutionSpecificationItem
from gaphor.UML.interactions.interaction import InteractionItem
from gaphor.UML.interactions.lifeline import LifelineItem
from gaphor.UML.interactions.message import MessageItem


def test_head_glue(diagram):
    """Test message head glue."""
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    # get head port
    port = ll.ports()[0]
    glued = allow(msg, msg.head, ll, port)
    assert glued


def test_visible_lifetime_glue(diagram):
    """Test message to visible lifetime glue."""
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    ll.lifetime.visible = True

    glued = allow(msg, msg.head, ll, ll.lifetime.port)
    assert glued


def test_lost_message_connection(diagram, element_factory):
    """Test lost message connection."""
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.head, ll)

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    # If one side is connected a "lost" message is created
    assert msg.subject is not None
    assert msg.subject.messageKind == "lost"

    assert len(messages) == 1
    assert len(occurrences) == 1
    assert messages[0] is msg.subject
    assert occurrences[0] is msg.subject.sendEvent


def test_found_message_connection(diagram, element_factory):
    """Test found message connection."""
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.tail, ll)

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    # If one side is connected a "found" message is created
    assert msg.subject is not None
    assert msg.subject.messageKind == "found"

    assert len(messages) == 1
    assert len(occurrences) == 1
    assert messages[0] is msg.subject
    assert occurrences[0] is msg.subject.receiveEvent


def test_complete_message_connection(diagram, element_factory):
    """Test complete message connection."""
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.head, ll1)
    connect(msg, msg.tail, ll2)

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    # two sides are connected - "complete" message is created
    assert msg.subject is not None
    assert msg.subject.messageKind == "complete"

    assert len(messages) == 1
    assert len(occurrences) == 2
    assert messages[0] is msg.subject
    assert msg.subject.sendEvent in occurrences, f"{occurrences}"
    assert msg.subject.receiveEvent in occurrences, f"{occurrences}"


def test_lifetime_connection(diagram):
    """Test messages' lifetimes connection."""
    msg = diagram.create(MessageItem)
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)

    # make lifelines to be in sequence diagram mode
    ll1.lifetime.visible = True
    ll2.lifetime.visible = True
    assert ll1.lifetime.visible and ll2.lifetime.visible

    # connect lifetimes with a message to the lifeline's head
    connect(msg, msg.head, ll1, ll1.lifetime.port)
    connect(msg, msg.tail, ll2, ll2.lifetime.port)

    assert msg.subject is not None
    assert msg.subject.messageKind == "complete"


def test_lifetime_connect_disconnect(diagram):
    """Test messages' lifetimes connection."""
    msg = diagram.create(MessageItem)
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)

    # make lifelines to be in sequence diagram mode
    ll1.lifetime.visible = True
    ll2.lifetime.visible = True
    assert ll1.lifetime.visible and ll2.lifetime.visible

    # connect lifetimes with a message to the lifeline's head
    connect(msg, msg.head, ll1, ll1.lifetime.port)
    connect(msg, msg.tail, ll2, ll2.lifetime.port)
    disconnect(msg, msg.tail)

    assert msg.subject is not None
    assert msg.subject.messageKind == "lost"


@pytest.mark.parametrize("end_name", ["head", "tail"])
def test_message_is_owned_by_implicit_interaction_connecting_to_head(
    diagram, element_factory, end_name
):
    """Test messages' lifetimes connection."""
    interaction = element_factory.create(UML.Interaction)
    msg = diagram.create(MessageItem)
    ll = diagram.create(LifelineItem, subject=element_factory.create(UML.Lifeline))
    ll.subject.interaction = interaction

    connect(msg, getattr(msg, end_name), ll)

    assert msg.subject is not None
    assert msg.subject.interaction is interaction
    assert msg.parent is None


@pytest.mark.parametrize("end_name", ["head", "tail"])
def test_message_is_owned_by_interaction_item_connecting_to_one_end(
    diagram, element_factory, end_name
):
    """Test messages' lifetimes connection."""
    interaction = diagram.create(
        InteractionItem, subject=element_factory.create(UML.Interaction)
    )
    ll = diagram.create(LifelineItem, subject=element_factory.create(UML.Lifeline))
    assert group(interaction.subject, ll.subject)

    msg = diagram.create(MessageItem)
    connect(msg, getattr(msg, end_name), ll)

    assert msg.subject is not None
    assert msg.subject.interaction is interaction.subject


def test_disconnection(diagram):
    """Test message disconnection."""
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.head, ll1)
    connect(msg, msg.tail, ll2)

    # one side disconnection
    disconnect(msg, msg.head)
    assert msg.subject is not None, f"{msg.subject}"

    # 2nd side disconnection
    disconnect(msg, msg.tail)
    assert msg.subject is None, f"{msg.subject}"


def test_communication_diagram_message_head_glue(diagram):
    communication = diagram.create(LifelineItem)
    sequence = diagram.create(LifelineItem)
    sequence.lifetime.visible = True
    message = diagram.create(MessageItem)

    connect(message, message.tail, sequence)

    assert allow(message, message.head, communication, communication.lifetime.port)


def test_communication_diagram_message_tail_glue(diagram):
    lifeline1 = diagram.create(LifelineItem)
    lifeline2 = diagram.create(LifelineItem)
    message = diagram.create(MessageItem)

    lifeline2.lifetime.visible = True
    connect(message, message.head, lifeline1)

    assert allow(message, message.tail, lifeline2, lifeline2.lifetime.port)


def test_message_glue_from_lifetimee_to_head(diagram):
    """Test gluing message on communication diagram."""

    lifeline1 = diagram.create(LifelineItem)
    lifeline2 = diagram.create(LifelineItem)
    message = diagram.create(MessageItem)

    lifeline1.lifetime.visible = True

    # connect head of message to lifeline's head
    connect(message, message.head, lifeline1)

    assert allow(message, message.tail, lifeline2, lifeline2.lifetime.port)


def test_messages_disconnect_cd(diagram, element_factory):
    """Test disconnecting messages on communication diagram."""
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.head, ll1)
    connect(msg, msg.tail, ll2)

    subject = msg.subject

    assert subject.sendEvent
    assert subject.receiveEvent

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    # verify integrity of messages
    assert len(messages) == 1
    assert len(occurrences) == 2


def test_message_connect_to_execution_specification(diagram, element_factory):
    """Test gluing message on sequence diagram."""

    lifeline = diagram.create(
        LifelineItem, subject=element_factory.create(UML.Lifeline)
    )
    exec_spec = diagram.create(ExecutionSpecificationItem)
    message = diagram.create(MessageItem)
    connect(exec_spec, exec_spec.handles()[0], lifeline, lifeline.lifetime.port)

    connect(message, message.head, exec_spec, exec_spec.ports()[0])

    assert message.subject
    assert message.subject.sendEvent.covered is lifeline.subject


def test_message_connect_to_execution_specification_in_interaction(
    diagram, element_factory
):
    interaction = diagram.create(
        InteractionItem, subject=element_factory.create(UML.Interaction)
    )

    lifeline = diagram.create(
        LifelineItem, subject=element_factory.create(UML.Lifeline)
    )
    assert group(interaction.subject, lifeline.subject)

    exec_spec = diagram.create(ExecutionSpecificationItem)
    connect(exec_spec, exec_spec.handles()[0], lifeline, lifeline.lifetime.port)

    message = diagram.create(MessageItem)
    connect(message, message.head, exec_spec, exec_spec.ports()[0])

    # In this particular case, the message is not (yet) tied to an interaction
    message.subject.interaction = None

    connect(message, message.tail, lifeline, lifeline.lifetime.port)

    assert lifeline.subject.interaction is interaction.subject
    assert exec_spec.subject.enclosingInteraction is interaction.subject
    assert message.subject.interaction is interaction.subject


def test_message_disconnect_from_execution_specification(
    diagram, element_factory, sanitizer_service
):
    """Test gluing message on sequence diagram."""

    lifeline = diagram.create(
        LifelineItem, subject=element_factory.create(UML.Lifeline)
    )
    exec_spec = diagram.create(ExecutionSpecificationItem)
    message = diagram.create(MessageItem)
    connect(exec_spec, exec_spec.handles()[0], lifeline, lifeline.lifetime.port)
    connect(message, message.head, exec_spec, exec_spec.ports()[0])

    disconnect(message, message.head)

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    assert not message.subject
    assert not len(messages)
    assert not len(occurrences)



import pytest

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.SysML import sysml
from gaphor.SysML.requirements.relationships import (
    DeriveReqtItem,
    RefineItem,
    SatisfyItem,
    TraceItem,
    VerifyItem,
)
from gaphor.SysML.requirements.requirement import RequirementItem
from gaphor.UML.classes import ContainmentItem


@pytest.mark.parametrize(
    "item_class", [DeriveReqtItem, RefineItem, SatisfyItem, TraceItem, VerifyItem]
)
def test_relation_allow_connect_disconnect_cycle(create, diagram, item_class):
    req1 = create(RequirementItem, sysml.Requirement)
    req2 = create(RequirementItem, sysml.Requirement)

    relation = diagram.create(item_class)

    assert allow(relation, relation.handles()[0], req1)
    assert allow(relation, relation.handles()[1], req2)

    connect(relation, relation.handles()[0], req1)
    connect(relation, relation.handles()[1], req2)

    assert relation.subject
    assert relation.subject.sourceContext is req1.subject
    assert relation.subject.targetContext is req2.subject

    disconnect(relation, relation.handles()[0])

    assert not relation.subject


def test_containment_between_requirements(create, diagram):
    """Test containment connecting between two requirements."""
    container_req = create(RequirementItem, sysml.Requirement)
    contained_req = create(RequirementItem, sysml.Requirement)
    line = create(ContainmentItem)

    connect(line, line.head, container_req)
    connect(line, line.tail, contained_req)

    assert diagram.connections.get_connection(line.tail).connected is contained_req
    assert len(container_req.subject.ownedElement) == 1
    assert contained_req.subject in container_req.subject.ownedElement


def test_containment_container_req_class_disconnect(create, diagram, element_factory):
    """Test containment disconnecting from two requirements."""
    parent_package = element_factory.create(UML.Package)
    diagram.element = parent_package
    container_req = create(RequirementItem, sysml.Requirement)
    contained_req = create(RequirementItem, sysml.Requirement)
    line = create(ContainmentItem)

    connect(line, line.tail, contained_req)
    connect(line, line.head, container_req)
    assert contained_req.subject not in parent_package.ownedElement

    disconnect(line, line.head)
    assert contained_req.subject in parent_package.ownedElement



"""Test general item connectors (Box, Ellipse, Line)."""

from gaphor.diagram.general import Box, Ellipse, Line
from gaphor.diagram.tests.fixtures import allow, connect, disconnect


def test_allow_line_to_box(create):
    box = create(Box)
    line = create(Line)

    allowed = allow(line, line.head, box)
    assert allowed


def test_connect_line_head_to_box(create):
    box = create(Box)
    line = create(Line)
    diagram = line.diagram

    connect(line, line.head, box)

    cinfo = diagram.connections.get_connection(line.head)
    assert cinfo.connected is box
    assert cinfo.item is line


def test_connect_line_tail_to_box(create):
    box = create(Box)
    line = create(Line)
    diagram = line.diagram

    connect(line, line.tail, box)

    cinfo = diagram.connections.get_connection(line.tail)
    assert cinfo.connected is box
    assert cinfo.item is line


def test_disconnect_line_from_box(create):
    box = create(Box)
    line = create(Line)
    diagram = line.diagram

    connect(line, line.head, box)
    assert diagram.connections.get_connection(line.head) is not None

    disconnect(line, line.head)
    assert diagram.connections.get_connection(line.head) is None


def test_connect_both_ends_to_boxes(create):
    box1 = create(Box)
    box2 = create(Box)
    line = create(Line)
    diagram = line.diagram

    connect(line, line.head, box1)
    connect(line, line.tail, box2)

    head_cinfo = diagram.connections.get_connection(line.head)
    tail_cinfo = diagram.connections.get_connection(line.tail)

    assert head_cinfo.connected is box1
    assert tail_cinfo.connected is box2
    assert head_cinfo.item is line
    assert tail_cinfo.item is line


def test_allow_line_to_ellipse(create):
    ellipse = create(Ellipse)
    line = create(Line)

    allowed = allow(line, line.head, ellipse)
    assert allowed


def test_connect_line_head_to_ellipse(create):
    ellipse = create(Ellipse)
    line = create(Line)
    diagram = line.diagram

    connect(line, line.head, ellipse)

    cinfo = diagram.connections.get_connection(line.head)
    assert cinfo.connected is ellipse
    assert cinfo.item is line


def test_connect_line_tail_to_ellipse(create):
    ellipse = create(Ellipse)
    line = create(Line)
    diagram = line.diagram

    connect(line, line.tail, ellipse)

    cinfo = diagram.connections.get_connection(line.tail)
    assert cinfo.connected is ellipse
    assert cinfo.item is line


def test_disconnect_line_from_ellipse(create):
    ellipse = create(Ellipse)
    line = create(Line)
    diagram = line.diagram

    connect(line, line.head, ellipse)
    assert diagram.connections.get_connection(line.head) is not None

    disconnect(line, line.head)
    assert diagram.connections.get_connection(line.head) is None


def test_connect_line_between_box_and_ellipse(create):
    box = create(Box)
    ellipse = create(Ellipse)
    line = create(Line)
    diagram = line.diagram

    connect(line, line.head, box)
    connect(line, line.tail, ellipse)

    head_cinfo = diagram.connections.get_connection(line.head)
    tail_cinfo = diagram.connections.get_connection(line.tail)

    assert head_cinfo.connected is box
    assert tail_cinfo.connected is ellipse



# ruff: noqa: F401,F811

import pytest

from gaphor import UML
from gaphor.UML import recipes
from gaphor.UML.classes.classeseditors import association_item_editor


@pytest.mark.asyncio
async def test_association_item_editor(diagram, element_factory, view, event_manager):
    type_a = element_factory.create(UML.Class)
    type_b = element_factory.create(UML.Class)
    item = diagram.create(
        UML.classes.AssociationItem, subject=recipes.create_association(type_a, type_b)
    )
    view.selection.hovered_item = item
    await view.update()
    result = association_item_editor(item, view, event_manager)

    assert result is True



import sys

import pytest

from gaphor.storage.xmlwriter import XMLWriter


class Writer:
    def __init__(self):
        self.s = ""

    def write(self, text):
        self.s += text


def test_elements_1():
    w = Writer()
    xml_w = XMLWriter(w)
    with xml_w.document():
        with xml_w.element("foo", {}):
            pass

    xml = f"""<?xml version="1.0" encoding="{sys.getdefaultencoding()}"?>\n<foo/>"""
    assert w.s == xml, f"{w.s} != {xml}"


def test_elements_2():
    w = Writer()
    xml_w = XMLWriter(w)
    with xml_w.document():
        with xml_w.element("foo", {}):
            with xml_w.element("bar", {}):
                pass

    xml = f"""<?xml version="1.0" encoding="{sys.getdefaultencoding()}"?>\n<foo>\n<bar/>\n</foo>"""
    assert w.s == xml, w.s


def test_elements_test():
    w = Writer()
    xml_w = XMLWriter(w)
    with xml_w.document():
        with xml_w.element("foo", {}):
            with xml_w.element("bar", {}):
                xml_w.characters("hello")

    xml = f"""<?xml version="1.0" encoding="{sys.getdefaultencoding()}"?>\n<foo>\n<bar>hello</bar>\n</foo>"""
    assert w.s == xml, w.s


def test_elements_ns_default():
    w = Writer()
    xml_w = XMLWriter(w)
    with xml_w.document():
        xml_w.prefix_mapping(None, "http://gaphor.devjavu.com/schema")
        with xml_w.element_ns(("http://gaphor.devjavu.com/schema", "foo"), {}):
            pass

    xml = f"""<?xml version="1.0" encoding="{sys.getdefaultencoding()}"?>\n<foo xmlns="http://gaphor.devjavu.com/schema"/>"""
    assert w.s == xml, w.s


def test_elements_ns_1():
    w = Writer()
    xml_w = XMLWriter(w)
    with xml_w.document():
        xml_w.prefix_mapping("g", "http://gaphor.devjavu.com/schema")
        with xml_w.element_ns(("http://gaphor.devjavu.com/schema", "foo"), {}):
            pass

    xml = f"""<?xml version="1.0" encoding="{sys.getdefaultencoding()}"?>\n<g:foo xmlns:g="http://gaphor.devjavu.com/schema"/>"""
    assert w.s == xml, w.s


def test_prefix_mappings():
    w = Writer()
    with XMLWriter(w).document() as xml_w:
        xml_w.prefix_mapping("example", "http://example.com/schema")
        with xml_w.element_ns(("http://example.com/schema", "foo"), {}):
            with xml_w.element_ns(("http://example.com/schema", "bar"), {}):
                pass

        with pytest.raises(KeyError):
            with xml_w.element_ns(("http://example.com/schema", "foo"), {}):
                pass



from gaphor.SysML import sysml


def test_requirement_traced_to(element_factory):
    req = element_factory.create(sysml.Requirement)
    traced = element_factory.create(sysml.Requirement)
    trace = element_factory.create(sysml.Trace)
    trace.sourceContext = req
    trace.targetContext = traced

    assert traced in req.tracedTo


def test_requirement_derived(element_factory):
    req = element_factory.create(sysml.Requirement)
    derived = element_factory.create(sysml.Requirement)
    trace = element_factory.create(sysml.DeriveReqt)
    trace.sourceContext = req
    trace.targetContext = derived

    assert derived in req.derived


def test_requirement_derived_from(element_factory):
    req = element_factory.create(sysml.Requirement)
    derived = element_factory.create(sysml.Requirement)
    trace = element_factory.create(sysml.DeriveReqt)
    trace.sourceContext = req
    trace.targetContext = derived

    assert req in derived.derivedFrom


def test_requirement_master(element_factory):
    req = element_factory.create(sysml.Requirement)
    copy_req = element_factory.create(sysml.Requirement)
    copy = element_factory.create(sysml.Copy)
    copy.sourceContext = copy_req
    copy.targetContext = req

    assert req in copy_req.master


def test_requirement_refined_by(element_factory):
    req = element_factory.create(sysml.Requirement)
    refined = element_factory.create(sysml.Requirement)
    refine = element_factory.create(sysml.Refine)
    refine.sourceContext = req
    refine.targetContext = refined

    assert refined in req.refinedBy


def test_requirement_satisfied_by(element_factory):
    req = element_factory.create(sysml.Requirement)
    satisfier = element_factory.create(sysml.Requirement)
    satisfy = element_factory.create(sysml.Satisfy)
    satisfy.sourceContext = req
    satisfy.targetContext = satisfier

    assert satisfier in req.satisfiedBy


def test_requirement_verified_by(element_factory):
    req = element_factory.create(sysml.Requirement)
    verifier = element_factory.create(sysml.Requirement)
    verify = element_factory.create(sysml.Verify)
    verify.sourceContext = req
    verify.targetContext = verifier

    assert verifier in req.verifiedBy



"""Formatting of UML model elements into text tests."""

import pytest

from gaphor.core.eventmanager import EventManager
from gaphor.core.format import format, parse
from gaphor.core.modeling import ElementFactory
from gaphor.UML import recipes
from gaphor.UML import uml as UML
from gaphor.UML.recipes import apply_stereotype
from gaphor.UML.umlfmt import format_association_end


@pytest.fixture
def factory():
    event_manager = EventManager()
    return ElementFactory(event_manager)


def add_tag_is_foo_metadata_field(e, factory):
    s = factory.create(UML.Stereotype)
    s.ownedAttribute = factory.create(UML.Property)
    parse(s.ownedAttribute[0], "tag: str")

    instance_spec = recipes.apply_stereotype(e, s)
    slot = recipes.add_slot(instance_spec, s.ownedAttribute[0])
    slotValue = factory.create(UML.LiteralString)
    slotValue.value = "foo"
    slot.value = slotValue
    return slot


@pytest.mark.parametrize(
    "text,formatted_text",
    [
        ("", ""),
        ("an attribute===foobar", "+ an attribute = ==foobar"),
        ("myattr", "+ myattr"),
        ("myattr:int", "+ myattr: int"),
        ("- myattr:int[3]", "- myattr: int[3]"),
        ("- myattr:int[0..1]", "- myattr: int[0..1]"),
        ("/myattr:int", "+ /myattr: int"),
        ("myattr:int=3", "+ myattr: int = 3"),
        ("myattr: int#some note", "+ myattr: int # some note"),
        ("# myattr:int=3 #some note", "# myattr: int = 3 # some note"),
        ("+ myattr: int | str", "+ myattr: int | str"),
    ],
)
def test_attribute(factory, text, formatted_text):
    """Test simple attribute formatting."""
    a = factory.create(UML.Property)
    parse(a, text)

    assert formatted_text == format(a, note=True)


def test_read_only_attribute(factory):
    a = factory.create(UML.Property)
    a.name = "myattr"
    a.isReadOnly = True

    assert "+ myattr { readOnly }" == format(a, tags=True)


def test_attribute_with_applied_stereotype(factory):
    a = factory.create(UML.Property)
    parse(a, "myattr: int")
    add_tag_is_foo_metadata_field(a, factory)

    assert '+ myattr: int { tag = "foo" }' == format(a, tags=True)


@pytest.mark.parametrize(
    "text,name_part, mult_part",
    [
        ("", "", ""),
        ("errorous (name)[3]", "+ errorous (name)[3]", ""),
        ("association end with spaces[3]", "+ association end with spaces", "3"),
        ("myattr", "+ myattr", ""),
        ("myattr[0..1]", "+ myattr", "0..1"),
        ("- myattr[0..1]", "- myattr", "0..1"),
    ],
)
def test_association_end(factory, text, name_part, mult_part):
    """Test simple attribute formatting."""
    a = factory.create(UML.Property)
    a.association = factory.create(UML.Association)
    parse(a, text)

    assert (name_part, mult_part) == format_association_end(a)


def test_attribute_with_type(factory):
    """Test simple attribute formatting."""
    a = factory.create(UML.Property)
    a.type = factory.create(UML.Class)
    a.name = "attr"
    a.type.name = "MyClass"

    assert "+ attr: MyClass" == format(a)


def test_association_end_with_applied_stereotype(factory):
    a = factory.create(UML.Property)
    a.association = factory.create(UML.Association)
    parse(a, "myattr[1]")
    add_tag_is_foo_metadata_field(a, factory)

    assert ("+ myattr", '1 { tag = "foo" }') == format_association_end(a)


@pytest.mark.parametrize(
    "text,formatted_text",
    [
        ("", ""),
        ("not an operation", "+ not an operation()"),
        ("+ myoper(param: str): int", "+ myoper(in param: str): int"),
        ("+ myoper(param: str = 'aap')", "+ myoper(in param: str = 'aap')"),
        ("- myoper(out param: str): int[2]", "- myoper(out param: str): int[2]"),
        ("- myoper(out param: str): int[0..3]", "- myoper(out param: str): int[0..3]"),
        ("- myoper(p1: str[2], p2:int[*])", "- myoper(in p1: str[2], in p2: int[*])"),
        (
            "- myoper(p1: str[2], p2:int[1..*])",
            "- myoper(in p1: str[2], in p2: int[1..*])",
        ),
        ("+ (param: str): int", "+ (param: str): int"),
        ("+ myoper(param: str): int#a note", "+ myoper(in param: str): int # a note"),
    ],
)
def test_operation(factory, text, formatted_text):
    """Test simple operation formatting."""
    o = factory.create(UML.Operation)
    parse(o, text)

    assert formatted_text == format(o, note=True)


def test_operation_with_stereotype(factory):
    o = factory.create(UML.Operation)
    st = factory.create(UML.Stereotype)
    st.name = "Stereotype"
    parse(o, "oper()")
    apply_stereotype(o, st)

    assert "+ «stereotype» oper()" == format(o, stereotype=True)


@pytest.mark.parametrize(
    "text,formatted_text",
    [
        ("", ""),
        ("param", "in param"),
        ("in param: str", "in param: str"),
        ("param = val", "in param = val"),
    ],
)
def test_parameter(factory, text, formatted_text):
    """Test simple operation formatting."""
    p = factory.create(UML.Parameter)
    parse(p, text)

    assert formatted_text == format(p)


def test_slot(factory):
    a = factory.create(UML.Property)
    parse(a, "myattr: int")
    slot = add_tag_is_foo_metadata_field(a, factory)

    assert 'tag = "foo"' == format(slot)


def test_pin(factory):
    pin = factory.create(UML.InputPin)
    pin.name = "foo"

    pin.type = factory.create(UML.Class)
    pin.type.name = "MyClass"

    recipes.set_multiplicity_lower_value(pin, 1)
    recipes.set_multiplicity_upper_value(pin, "*")

    assert format(pin) == "foo: MyClass[1..*]"


@pytest.mark.parametrize(
    "text,formatted_text",
    [
        ("", ""),
        ("param", "in param"),
        ("in param: str", "in param: str"),
        ("param = val", "in param = val"),
    ],
)
def test_activity_parameter_node(factory, text, formatted_text):
    """Test simple operation formatting."""
    p = factory.create(UML.Parameter)
    n = factory.create(UML.ActivityParameterNode)
    n.parameter = p
    parse(p, text)

    assert formatted_text == format(n)
